from __future__ import division, print_function

import os
import subprocess
import sys
from collections import namedtuple

import pytest

from leapp import models, reporting
from leapp.exceptions import StopActorExecution, StopActorExecutionError
from leapp.libraries.actor import userspacegen
from leapp.libraries.common import overlaygen, repofileutils, rhsm
from leapp.libraries.common.config import architecture
from leapp.libraries.common.testutils import CurrentActorMocked, logger_mocked, produce_mocked
from leapp.utils.deprecation import suppress_deprecation

if sys.version_info < (2, 8):
    from pathlib2 import Path
else:
    from pathlib import Path


CUR_DIR = os.path.dirname(os.path.abspath(__file__))
_CERTS_PATH = os.path.join(CUR_DIR, '../../../files', userspacegen.PROD_CERTS_FOLDER)
_DEFAULT_CERT_PATH = os.path.join(_CERTS_PATH, '8.1', '479.pem')


@pytest.fixture
def adjust_cwd():
    previous_cwd = os.getcwd()
    os.chdir(os.path.join(CUR_DIR, "../"))
    yield
    os.chdir(previous_cwd)


class MockedMountingBase(object):
    def __init__(self, **dummy_kwargs):
        self.called_copytree_from = []
        self.target = ''

    def copytree_from(self, src, dst):
        self.called_copytree_from.append((src, dst))

    def __call__(self, **dummy_kwarg):
        yield self

    def call(self, *args, **kwargs):
        return {'stdout': ''}

    def nspawn(self):
        return self

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        pass


def traverse_structure(structure, root=Path('/')):
    """
    Given a description of a directory structure, return fullpaths to the
    files and what they link to.

    :param structure: A dict which defined the directory structure.  See below
        for what it looks like.
    :param root: A path to prefix to the files.  On an actual run in production.
        this would be `/` but since we're doing this in a unittest, it needs to
        be a temporary directory.
    :returns: This is a generator, so pairs of (filepath, what it links to) will
        be returned one at a time, each time through the iterable.

    The semantics of `structure` are as follows:

    1. The outermost dictionary encodes the root of a directory structure

    2. Depending on the value for a key in a dict, each key in the dictionary
       denotes the name of either a:
         a) directory -- if value is dict
         b) regular file -- if value is None
         c) symlink -- if a value is str

     3. The value of a symlink entry is a absolute path to a file in the context of
        the structure.

    .. warning:: Empty directories are not returned.
    """
    for filename, links_to in structure.items():
        filepath = root / filename

        if isinstance(links_to, dict):
            for pair in traverse_structure(links_to, filepath):
                yield pair
        else:
            yield (filepath, links_to)


def assert_directory_structure_matches(root, initial, expected):
    # Assert every file that is supposed to be present is present
    for filepath, links_to in traverse_structure(expected, root=root / 'expected'):
        assert filepath.exists(), "{} was supposed to exist and does not".format(filepath)

        if links_to is None:
            assert filepath.is_file(), "{} was supposed to be a file but is not".format(filepath)
            continue

        assert filepath.is_symlink(), '{} was supposed to be a symlink but is not'.format(filepath)

        # We need to rewrite absolute paths because:
        # * links_to contains an absolute path to the resource where the root
        #   directory is `/`.
        # * In our test case, the source resource is rooted in a temporary
        #   directory rather than '/'.
        # * The temporary directory name is root / 'initial'.
        # So we rewrite the initial `/` to be `root/{initial}` to account for
        # that.  In production, the root directory will be `/` so no rewriting
        # will happen there.
        #
        if links_to.startswith('/'):
            links_to = str(root / 'initial' / links_to.lstrip('/'))

        actual_links_to = os.readlink(str(filepath))
        assert actual_links_to == str(links_to), (
            '{} linked to {} instead of {}'.format(filepath, actual_links_to, links_to))

    # Assert there are no extra files
    result_dir = str(root / 'expected')
    for fileroot, dummy_dirs, files in os.walk(result_dir):
        for filename in files:
            dir_path = os.path.relpath(fileroot, result_dir).split('/')

            cwd = expected
            for directory in dir_path:
                cwd = cwd[directory]

            assert filename in cwd

            filepath = os.path.join(fileroot, filename)
            if os.path.islink(filepath):
                links_to = os.readlink(filepath)
                # We rewrite absolute paths because the root directory is in
                # a temp dir instead of `/` in the unittest.  See the comment
                # where we rewrite `links_to` for the previous loop in this
                # function for complete details.
                if links_to.startswith('/'):
                    links_to = '/' + os.path.relpath(links_to, str(root / 'initial'))
                assert cwd[filename] == links_to


@pytest.fixture
def temp_directory_layout(tmp_path, initial_structure):
    for filepath, links_to in traverse_structure(initial_structure, root=tmp_path / 'initial'):
        # Directories are inlined by traverse_structure so we need to create
        # them here
        file_path = tmp_path / filepath
        file_path.parent.mkdir(parents=True, exist_ok=True)

        # Real file
        if links_to is None:
            file_path.touch()
            continue

        # Symlinks
        if links_to.startswith('/'):
            # Absolute symlink
            file_path.symlink_to(tmp_path / 'initial' / links_to.lstrip('/'))
        else:
            # Relative symlink
            file_path.symlink_to(links_to)

    (tmp_path / 'expected').mkdir()
    assert (tmp_path / 'expected').exists()

    return tmp_path


# The semantics of initial_structure and expected_structure are defined in the
# traverse_structure() docstring.
@pytest.mark.parametrize('initial_structure,expected_structure', [
    (pytest.param(
        {
            'dir': {
                'fileA': None
            }
        },
        {
            'dir': {
                'fileA': None
            },
        },
        id="Copy_a_regular_file"
    )),
    # Absolute symlink tests
    (pytest.param(
        {
            'dir': {
                'fileA': '/nonexistent'
            }
        },
        {
            'dir': {},
        },
        id="Absolute_do_not_copy_a_broken_symlink"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': '/dir/fileB',
                'fileB': '/nonexistent'
            }
        },
        {
            'dir': {}
        },
        id="Absolute_do_not_copy_a_chain_of_broken_symlinks"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': '/nonexistent-dir/nonexistent'
            },
        },
        {
            'dir': {},
        },
        id="Absolute_do_not_copy_a_broken_symlink_to_a_nonexistent_directory"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': '/dir/fileB',
                'fileB': '/dir/fileC',
                'fileC': '/dir/fileA',
                'fileD': '/dir/fileD',
            }
        },
        {
            'dir': {}
        },
        id="Absolute_do_not_copy_circular_symlinks"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': '/dir/fileB',
                'fileB': None
            }
        },
        {
            'dir': {
                'fileA': '/dir/fileB',
                'fileB': None
            }
        },
        id="Absolute_copy_a_regular_symlink"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': '/dir/fileB',
                'fileB': '/dir/fileC',
                'fileC': None
            }
        },
        {
            'dir': {
                'fileA': '/dir/fileB',
                'fileB': '/dir/fileC',
                'fileC': None
            }
        },
        id="Absolute_copy_a_chain_of_symlinks"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': '/dir/fileB',
                'fileB': '/dir/fileC',
                'fileC': '/outside/fileOut',
                'fileE': None
            },
            'outside': {
                'fileOut': '/outside/fileD',
                'fileD': '/dir/fileE'
            }
        },
        {
            'dir': {
                'fileA': '/dir/fileB',
                'fileB': '/dir/fileC',
                'fileC': '/dir/fileE',
                'fileE': None,
            }
        },
        id="Absolute_copy_a_link_to_a_file_outside_the_considered_directory_as_file"
    )),
    (pytest.param(
        {
            'dir': {
                'nested': {
                    'fileA': '/dir/nested/fileB',
                    'fileB': '/dir/nested/fileC',
                    'fileC': '/outside/fileOut',
                    'fileE': None
                }
            },
            'outside': {
                'fileOut': '/outside/fileD',
                'fileD': '/dir/nested/fileE'
            }
        },
        {
            'dir': {
                'nested': {
                    'fileA': '/dir/nested/fileB',
                    'fileB': '/dir/nested/fileC',
                    'fileC': '/dir/nested/fileE',
                    'fileE': None
                }
            }
        },
        id="Absolute_copy_a_link_to_a_file_outside_with_a_nested_structure_within_the_source_dir"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': '/dir/fileB',
                'fileB': '/dir/fileC',
                'fileC': '/outside/nested/fileOut',
                'fileE': None
            },
            'outside': {
                'nested': {
                    'fileOut': '/outside/nested/fileD',
                    'fileD': '/dir/fileE'
                }
            }
        },
        {
            'dir': {
                'fileA': '/dir/fileB',
                'fileB': '/dir/fileC',
                'fileC': '/dir/fileE',
                'fileE': None,
            }
        },
        id="Absolute_copy_a_link_to_a_file_outside_with_a_nested_structure_in_the_outside_dir"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': '/outside/fileOut',
                'fileB': None,
            },
            'outside': {
                'fileOut': '../dir/fileB',
            },
        },
        {
            'dir': {
                'fileA': '/dir/fileB',
                'fileB': None,
            },
        },
        id="Absolute_symlink_that_leaves_the_directory_but_returns_with_relative_outside"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': '/outside/fileB',
                'fileB': None,
            },
            'outside': '/dir',
        },
        {
            'dir': {
                'fileA': '/dir/fileB',
                'fileB': None,
            },
        },
        id="Absolute_symlink_to_a_file_inside_via_a_symlink_to_the_rootdir"
    )),
    # This should be fixed but not necessarily for this release.
    # It makes sure that when we have two separate links to the
    # same file outside of /etc/pki, one of the links is copied
    # as a real file and the other is made a link to the copy.
    # (Right now, the real file is copied in place of both links.)
    # (pytest.param(
    #     {
    #         'dir': {
    #             'fileA': '/outside/fileC',
    #             'fileB': '/outside/fileC',
    #         },
    #         'outside': {
    #             'fileC': None,
    #         },
    #     },
    #     {
    #         'dir': {
    #             'fileA': None,
    #             'fileB': '/dir/fileA',
    #         },
    #     },
    #     id="Absolute_two_symlinks_to_the_same_copied_file"
    # )),
    (pytest.param(
        {
            'dir': {
                'fileA': None,
                'link_to_dir': '/dir/inside',
                'inside': {
                    'fileB': None,
                },
            },
        },
        {
            'dir': {
                'fileA': None,
                'link_to_dir': '/dir/inside',
                'inside': {
                    'fileB': None,
                },
            },
        },
        id="Absolute_symlink_to_a_dir_inside"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': None,
                'link_to_dir': '/outside',
            },
            'outside': {
                'fileB': None,
            },
        },
        {
            'dir': {
                'fileA': None,
                'link_to_dir': {
                    'fileB': None,
                },
            },
        },
        id="Absolute_symlink_to_a_dir_outside"
    )),
    (pytest.param(
        # This one is very tricky:
        # * The user has made /etc/pki a symlink to some other directory that
        #   they keep certificates.
        # * In the target system, we are going to make /etc/pki an actual
        #   directory with the contents that the actual directory on the host
        #   system had.
        {
            'dir': '/funkydir',
            'funkydir': {
                'fileA': '/funkydir/fileB',
                'fileB': None,
            },
        },
        {
            'dir': {
                'fileA': '/dir/fileB',
                'fileB': None,
            },
        },
        id="Absolute_symlink_where_srcdir_is_a_symlink_on_the_host_system"
    )),
    # Relative symlink tests
    (pytest.param(
        {
            'dir': {
                'fileA': 'nonexistent'
            },
        },
        {
            'dir': {},
        },
        id="Relative_do_not_copy_a_broken_symlink"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': 'nonexistent'
            }
        },
        {
            'dir': {}
        },
        id="Relative_do_not_copy_a_chain_of_broken_symlinks"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': 'nonexistent-dir/nonexistent'
            },
        },
        {
            'dir': {},
        },
        id="Relative_do_not_copy_a_broken_symlink_to_a_nonexistent_directory"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': 'fileC',
                'fileC': 'fileA',
                'fileD': 'fileD',
            }
        },
        {
            'dir': {}
        },
        id="Relative_do_not_copy_circular_symlinks"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': None,
            },
        },
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': None,
            },
        },
        id="Relative_copy_a_regular_symlink_to_a_file_in_the_same_directory"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': 'dir2/../fileB',
                'fileB': None,
                'dir2': {
                    'fileC': None
                },
            },
        },
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': None,
                'dir2': {
                    'fileC': None
                },
            },
        },
        id="Relative_symlink_with_parent_dir_but_still_in_same_directory"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': 'fileC',
                'fileC': None
            }
        },
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': 'fileC',
                'fileC': None
            }
        },
        id="Relative_copy_a_chain_of_symlinks"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': 'fileC',
                'fileC': '../outside/fileOut',
                'fileE': None
            },
            'outside': {
                'fileOut': 'fileD',
                'fileD': '../dir/fileE'
            }
        },
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': 'fileC',
                'fileC': 'fileE',
                'fileE': None,
            }
        },
        id="Relative_copy_a_link_to_a_file_outside_the_considered_directory_as_file"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': '../outside/fileOut',
                'fileB': None,
            },
            'outside': {
                'fileOut': None,
            },
        },
        {
            'dir': {
                'fileA': None,
                'fileB': None,
            },
        },
        id="Relative_symlink_to_outside"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': 'nested/fileB',
                'nested': {
                    'fileB': None,
                },
            },
        },
        {
            'dir': {
                'fileA': 'nested/fileB',
                'nested': {
                    'fileB': None,
                },
            },
        },
        id="Relative_copy_a_symlink_to_a_file_in_a_subdir"
    )),
    (pytest.param(
        {
            'dir': {
                'fileF': 'nested/fileC',
                'nested': {
                    'fileA': 'fileB',
                    'fileB': 'fileC',
                    'fileC': '../../outside/fileOut',
                    'fileE': None,
                }
            },
            'outside': {
                'fileOut': 'fileD',
                'fileD': '../dir/nested/fileE',
            }
        },
        {
            'dir': {
                'fileF': 'nested/fileC',
                'nested': {
                    'fileA': 'fileB',
                    'fileB': 'fileC',
                    'fileC': 'fileE',
                    'fileE': None,
                }
            }
        },
        id="Relative_copy_a_link_to_a_file_outside_with_a_nested_structure_within_the_source_dir"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': 'fileC',
                'fileC': '../outside/nested/fileOut',
                'fileE': None
            },
            'outside': {
                'nested': {
                    'fileOut': 'fileD',
                    'fileD': '../../dir/fileE'
                }
            }
        },
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': 'fileC',
                'fileC': 'fileE',
                'fileE': None,
            }
        },
        id="Relative_copy_a_link_to_a_file_outside_with_a_nested_structure_in_the_outside_dir"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': '../outside/fileOut',
                'fileB': None,
            },
            'outside': {
                'fileOut': '../dir/fileB',
            },
        },
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': None,
            },
        },
        id="Relative_symlink_that_leaves_the_directory_but_returns"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': '../outside/fileOut',
                'fileB': None,
            },
            'outside': {
                'fileOut': '/dir/fileB',
            },
        },
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': None,
            },
        },
        id="Relative_symlink_that_leaves_the_directory_but_returns_with_absolute_outside"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': '../outside/fileB',
                'fileB': None,
            },
            'outside': '/dir',
        },
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': None,
            },
        },
        id="Relative_symlink_to_a_file_inside_via_a_symlink_to_the_rootdir"
    )),
    # This should be fixed but not necessarily for this release.
    # It makes sure that when we have two separate links to the
    # same file outside of /etc/pki, one of the links is copied
    # as a real file and the other is made a link to the copy.
    # (Right now, the real file is copied in place of both links.)
    # (pytest.param(
    #     {
    #         'dir': {
    #             'fileA': '../outside/fileC',
    #             'fileB': '../outside/fileC',
    #         },
    #         'outside': {
    #             'fileC': None,
    #         },
    #     },
    #     {
    #         'dir': {
    #             'fileA': None,
    #             'fileB': 'fileA',
    #         },
    #     },
    #     id="Relative_two_symlinks_to_the_same_copied_file"
    # )),
    (pytest.param(
        {
            'dir': {
                'fileA': None,
                'link_to_dir': '../outside',
            },
            'outside': {
                'fileB': None,
            },
        },
        {
            'dir': {
                'fileA': None,
                'link_to_dir': {
                    'fileB': None,
                },
            },
        },
        id="Relative_symlink_to_a_dir_outside"
    )),
    (pytest.param(
        {
            'dir': {
                'fileA': None,
                'link_to_dir': 'inside',
                'inside': {
                    'fileB': None,
                },
            },
        },
        {
            'dir': {
                'fileA': None,
                'link_to_dir': 'inside',
                'inside': {
                    'fileB': None,
                },
            },
        },
        id="Relative_symlink_to_a_dir_inside"
    )),
    (pytest.param(
        # This one is very tricky:
        # * The user has made /etc/pki a symlink to some other directory that
        #   they keep certificates.
        # * In the target system, we are going to make /etc/pki an actual
        #   directory with the contents that the actual directory on the host
        #   system had.
        {
            'dir': 'funkydir',
            'funkydir': {
                'fileA': 'fileB',
                'fileB': None,
            },
        },
        {
            'dir': {
                'fileA': 'fileB',
                'fileB': None,
            },
        },
        id="Relative_symlink_where_srcdir_is_a_symlink_on_the_host_system"
    )),
]
)
def test_copy_decouple(monkeypatch, temp_directory_layout, initial_structure, expected_structure):

    def run_mocked(command):
        subprocess.check_call(
            ' '.join(command),
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

    monkeypatch.setattr(userspacegen, 'run', run_mocked)
    expected_dir = temp_directory_layout / 'expected' / 'dir'
    expected_dir.mkdir()
    userspacegen._copy_decouple(
            str(temp_directory_layout / 'initial' / 'dir'),
            str(expected_dir),
            )

    try:
        assert_directory_structure_matches(temp_directory_layout, initial_structure, expected_structure)
    except AssertionError:
        # For debugging purposes, print out the entire directory structure if an
        # assertion failed.
        for rootdir, dirs, files in os.walk(temp_directory_layout):
            for d in dirs:
                print(os.path.join(rootdir, d))
            for f in files:
                filename = os.path.join(rootdir, f)
                print("  {}".format(filename))
                if os.path.islink(filename):
                    print("    => Links to: {}".format(os.readlink(filename)))

        # Then re-raise the assertion
        raise


@pytest.mark.parametrize('result,dst_ver,arch,prod_type', [
    (os.path.join(_CERTS_PATH, '8.1', '479.pem'), '8.1', architecture.ARCH_X86_64, 'ga'),
    (os.path.join(_CERTS_PATH, '8.1', '419.pem'), '8.1', architecture.ARCH_ARM64, 'ga'),
    (os.path.join(_CERTS_PATH, '8.1', '279.pem'), '8.1', architecture.ARCH_PPC64LE, 'ga'),
    (os.path.join(_CERTS_PATH, '8.2', '479.pem'), '8.2', architecture.ARCH_X86_64, 'ga'),
    (os.path.join(_CERTS_PATH, '8.5', '486.pem'), '8.5', architecture.ARCH_X86_64, 'beta'),
    (os.path.join(_CERTS_PATH, '8.2', '72.pem'), '8.2', architecture.ARCH_S390X, 'ga'),
    (os.path.join(_CERTS_PATH, '8.5', '433.pem'), '8.5', architecture.ARCH_S390X, 'beta'),
])
def test_get_product_certificate_path(monkeypatch, adjust_cwd, result, dst_ver, arch, prod_type):
    envars = {'LEAPP_DEVEL_TARGET_PRODUCT_TYPE': prod_type}
    curr_actor_mocked = CurrentActorMocked(dst_ver=dst_ver, arch=arch, envars=envars)
    monkeypatch.setattr(userspacegen.api, 'current_actor', curr_actor_mocked)
    assert userspacegen._get_product_certificate_path() in result


@suppress_deprecation(models.RequiredTargetUserspacePackages)
def _gen_packages_msgs():
    _cfiles = [
        models.CopyFile(src='/path/src', dst='/path/dst'),
        models.CopyFile(src='/path/foo', dst='/path/bar'),
    ]
    return [
        models.RequiredTargetUserspacePackages(),
        models.RequiredTargetUserspacePackages(packages=['pkgA']),
        models.RequiredTargetUserspacePackages(packages=['pkgB', 'pkgsC']),
        models.RequiredTargetUserspacePackages(packages=['pkgD']),
        models.TargetUserSpacePreupgradeTasks(),
        models.TargetUserSpacePreupgradeTasks(install_rpms=['pkgA']),
        models.TargetUserSpacePreupgradeTasks(install_rpms=['pkgB', 'pkgsC']),
        models.TargetUserSpacePreupgradeTasks(install_rpms=['pkgD', 'pkgE'], copy_files=[_cfiles[0]]),
        models.TargetUserSpacePreupgradeTasks(copy_files=_cfiles),
    ]


_PACKAGES_MSGS = _gen_packages_msgs()
_RHSMINFO_MSG = models.RHSMInfo(attached_skus=['testing-sku'])
_RHUIINFO_MSG = models.RHUIInfo(provider='aws',
                                src_client_pkg_names=['rh-amazon-rhui-client'],
                                target_client_pkg_names=['rh-amazon-rhui-client'],
                                target_client_setup_info=models.TargetRHUISetupInfo(
                                    preinstall_tasks=models.TargetRHUIPreInstallTasks(),
                                    postinstall_tasks=models.TargetRHUIPostInstallTasks()))
_XFS_MSG = models.XFSPresence()
_STORAGEINFO_MSG = models.StorageInfo()
_CTRF_MSGS = [
    models.CustomTargetRepositoryFile(file='rfileA'),
    models.CustomTargetRepositoryFile(file='rfileB'),
]
_SAEE = StopActorExecutionError
_SAE = StopActorExecution


class MockedConsume(object):
    def __init__(self, *args):
        self._msgs = []
        for arg in args:
            if not arg:
                continue
            if isinstance(arg, list):
                self._msgs.extend(arg)
            else:
                self._msgs.append(arg)

    def __call__(self, model):
        return iter([msg for msg in self._msgs if isinstance(msg, model)])


testInData = namedtuple(
    'TestInData', ['pkg_msgs', 'rhsm_info', 'rhui_info', 'xfs', 'storage', 'custom_repofiles']
)


# NOTE: tests cover know new, deprecated, and both ways how to require packages
# that should be installed to create the target userspace. Cases which could be
# removed completely after the drop of the deprecated functionality, are marked
# with the `# dep` str.
@pytest.mark.parametrize('raised,no_rhsm,testdata', [
    # valid cases with RHSM
    (None, '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    (None, '0', testInData(_PACKAGES_MSGS[:4], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None)),   # dep
    (None, '0', testInData(_PACKAGES_MSGS[4:8], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None)),  # dep
    (None, '0', testInData(_PACKAGES_MSGS[0], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    (None, '0', testInData(_PACKAGES_MSGS[4], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None)),    # dep
    (None, '0', testInData([], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    (None, '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, None, _STORAGEINFO_MSG, None)),
    (None, '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, _CTRF_MSGS)),
    (None, '0', testInData(_PACKAGES_MSGS[0], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, _CTRF_MSGS)),
    (None, '0', testInData(_PACKAGES_MSGS[4], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, _CTRF_MSGS)),  # dep
    (None, '0', testInData([], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, _CTRF_MSGS)),
    (None, '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, None, _STORAGEINFO_MSG, _CTRF_MSGS)),

    # valid cases without RHSM (== skip_rhsm)
    (None, '1', testInData(_PACKAGES_MSGS, None, _RHUIINFO_MSG, _XFS_MSG, _STORAGEINFO_MSG, None)),
    (None, '1', testInData(_PACKAGES_MSGS[:4], None, _RHUIINFO_MSG, _XFS_MSG, _STORAGEINFO_MSG, None)),   # dep
    (None, '1', testInData(_PACKAGES_MSGS[4:8], None, _RHUIINFO_MSG, _XFS_MSG, _STORAGEINFO_MSG, None)),  # dep
    (None, '1', testInData(_PACKAGES_MSGS, None, _RHUIINFO_MSG, None, _STORAGEINFO_MSG, None)),
    (None, '1', testInData([], None, _RHUIINFO_MSG, _XFS_MSG, _STORAGEINFO_MSG, None)),
    (None, '1', testInData([], None, _RHUIINFO_MSG, None, _STORAGEINFO_MSG, None)),
    (None, '1', testInData(_PACKAGES_MSGS, None, _RHUIINFO_MSG, _XFS_MSG, _STORAGEINFO_MSG, _CTRF_MSGS)),
    (None, '1', testInData(_PACKAGES_MSGS, None, _RHUIINFO_MSG, None, _STORAGEINFO_MSG, _CTRF_MSGS)),
    (None, '1', testInData([], None, _RHUIINFO_MSG, _XFS_MSG, _STORAGEINFO_MSG, _CTRF_MSGS)),
    (None, '1', testInData([], None, _RHUIINFO_MSG, None, _STORAGEINFO_MSG, _CTRF_MSGS)),

    # no-rhsm but RHSMInfo defined (should be _RHSMINFO_MSG)
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS[0], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG,
                                             None)),
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS[4], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG,
                                             None)),  # dep
    ((_SAEE, 'RHSM is not'), '1', testInData([], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, None, _STORAGEINFO_MSG, None)),
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG,
                                             _CTRF_MSGS)),
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS[0], _RHSMINFO_MSG, None, _XFS_MSG,
                                             _STORAGEINFO_MSG, _CTRF_MSGS)),
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS[4], _RHSMINFO_MSG, None, _XFS_MSG,
                                             _STORAGEINFO_MSG, _CTRF_MSGS)),  # dep
    ((_SAEE, 'RHSM is not'), '1', testInData([], _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, _CTRF_MSGS)),
    ((_SAEE, 'RHSM is not'), '1', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, None, _STORAGEINFO_MSG, _CTRF_MSGS)),

    # missing RHSMInfo but it should exist
    # NOTE: should be this Error?!
    ((_SAE, 'RHSM information'), '0', testInData(_PACKAGES_MSGS, None, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    ((_SAE, 'RHSM information'), '0', testInData(_PACKAGES_MSGS, None, None, None, _STORAGEINFO_MSG, None)),
    ((_SAE, 'RHSM information'), '0', testInData([], None, None, _XFS_MSG, _STORAGEINFO_MSG, None)),
    ((_SAE, 'RHSM information'), '0', testInData([], None, None, None, _STORAGEINFO_MSG, None)),

    # in the end, error when StorageInfo is missing
    ((_SAEE, 'No storage'), '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, _XFS_MSG, None, None)),
    ((_SAEE, 'No storage'), '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, None, None, None)),
    ((_SAEE, 'No storage'), '0', testInData([], _RHSMINFO_MSG, None, _XFS_MSG, None, None)),
    ((_SAEE, 'No storage'), '0', testInData([], _RHSMINFO_MSG, None, None, None, None)),
    ((_SAEE, 'No storage'), '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, _XFS_MSG, None, _CTRF_MSGS)),
    ((_SAEE, 'No storage'), '0', testInData(_PACKAGES_MSGS, _RHSMINFO_MSG, None, None, None, _CTRF_MSGS)),
    ((_SAEE, 'No storage'), '0', testInData([], _RHSMINFO_MSG, None, _XFS_MSG, None, _CTRF_MSGS)),
    ((_SAEE, 'No storage'), '0', testInData([], _RHSMINFO_MSG, None, None, None, _CTRF_MSGS)),
])
def test_consume_data(monkeypatch, raised, no_rhsm, testdata):
    # do not write never into testdata inside the test !!
    xfs = testdata.xfs
    custom_repofiles = testdata.custom_repofiles
    _exp_pkgs = {'dnf', 'dnf-command(config-manager)'}
    _exp_files = []

    def _get_pkgs(msg):
        if isinstance(msg, models.TargetUserSpacePreupgradeTasks):
            return msg.install_rpms
        return msg.packages

    def _get_files(msg):
        if isinstance(msg, models.TargetUserSpacePreupgradeTasks):
            return msg.copy_files
        return []

    def _cfiles2set(cfiles):
        return {(i.src, i.dst) for i in cfiles}

    if isinstance(testdata.pkg_msgs, list):
        for msg in testdata.pkg_msgs:
            _exp_pkgs.update(_get_pkgs(msg))
            _exp_files += _get_files(msg)
    else:
        _exp_pkgs.update(_get_pkgs(testdata.pkg_msgs))
        _exp_files += _get_files(testdata.pkg_msgs)
    mocked_consume = MockedConsume(testdata.pkg_msgs,
                                   testdata.rhsm_info,
                                   testdata.rhui_info,
                                   xfs,
                                   testdata.storage,
                                   custom_repofiles)

    monkeypatch.setattr(userspacegen.api, 'consume', mocked_consume)
    monkeypatch.setattr(userspacegen.api, 'current_logger', logger_mocked())
    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked(envars={'LEAPP_NO_RHSM': no_rhsm}))
    if not xfs:
        xfs = models.XFSPresence()
    if not custom_repofiles:
        custom_repofiles = []
    if not raised:
        result = userspacegen._InputData()
        assert result.packages == _exp_pkgs
        assert _cfiles2set(result.files) == _cfiles2set(_exp_files)
        assert result.rhsm_info == testdata.rhsm_info
        assert result.rhui_info == testdata.rhui_info
        assert result.xfs_info == xfs
        assert result.storage_info == testdata.storage
        assert result.custom_repofiles == custom_repofiles
        assert not userspacegen.api.current_logger.warnmsg
        assert not userspacegen.api.current_logger.errmsg
    else:
        with pytest.raises(raised[0]) as err:
            userspacegen._InputData()
        if isinstance(err.value, StopActorExecutionError):
            assert raised[1] in err.value.message
        else:
            assert userspacegen.api.current_logger.warnmsg
            assert any([raised[1] in x for x in userspacegen.api.current_logger.warnmsg])


@pytest.mark.skip(reason="Currently not implemented in the actor. It's TODO.")
def test_gather_target_repositories(monkeypatch):
    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked())
    # The available RHSM repos
    monkeypatch.setattr(rhsm, 'get_available_repo_ids', lambda x: ['repoidX', 'repoidY', 'repoidZ'])
    monkeypatch.setattr(rhsm, 'skip_rhsm', lambda: False)
    # The required RHEL repos based on the repo mapping and PES data + custom repos required by third party actors
    monkeypatch.setattr(userspacegen.api, 'consume', lambda x: iter([models.TargetRepositories(
        rhel_repos=[models.RHELTargetRepository(repoid='repoidX'),
                    models.RHELTargetRepository(repoid='repoidY')],
        custom_repos=[models.CustomTargetRepository(repoid='repoidCustom')])]))

    target_repoids = userspacegen.gather_target_repositories(None, None)

    assert target_repoids == ['repoidX', 'repoidY', 'repoidCustom']


def test_gather_target_repositories_none_available(monkeypatch):

    mocked_produce = produce_mocked()
    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(userspacegen.api.current_actor(), 'produce', mocked_produce)
    monkeypatch.setattr(rhsm, 'get_available_repo_ids', lambda x: [])
    monkeypatch.setattr(rhsm, 'skip_rhsm', lambda: False)
    with pytest.raises(StopActorExecution):
        userspacegen.gather_target_repositories(None, None)
        assert mocked_produce.called
        reports = [m.report for m in mocked_produce.model_instances if isinstance(m, reporting.Report)]
        inhibitors = [m for m in reports if 'INHIBITOR' in m.get('flags', ())]
        assert len(inhibitors) == 1
        assert inhibitors[0].get('title', '') == 'Cannot find required basic RHEL target repositories.'


def test_gather_target_repositories_rhui(monkeypatch):

    indata = testInData(
        _PACKAGES_MSGS, _RHSMINFO_MSG, _RHUIINFO_MSG, _XFS_MSG, _STORAGEINFO_MSG, None
    )

    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(userspacegen, '_get_all_available_repoids', lambda x: [])
    monkeypatch.setattr(
        userspacegen, '_get_rh_available_repoids', lambda x, y: ['rhui-1', 'rhui-2', 'rhui-3']
    )
    monkeypatch.setattr(rhsm, 'skip_rhsm', lambda: True)
    monkeypatch.setattr(
        userspacegen.api, 'consume', lambda x: iter(
            [models.TargetRepositories(
                rhel_repos=[
                    models.RHELTargetRepository(repoid='rhui-1'),
                    models.RHELTargetRepository(repoid='rhui-2')
                ]
            )
            ])
    )
    target_repoids = userspacegen.gather_target_repositories(None, indata)
    assert target_repoids == set(['rhui-1', 'rhui-2'])


def test_gather_target_repositories_baseos_appstream_not_available(monkeypatch):
    # If the repos that Leapp identifies as required for the upgrade (based on the repo mapping and PES data) are not
    # available, an exception shall be raised

    indata = testInData(
        _PACKAGES_MSGS, _RHSMINFO_MSG, None, _XFS_MSG, _STORAGEINFO_MSG, None
    )
    monkeypatch.setattr(rhsm, 'skip_rhsm', lambda: False)

    mocked_produce = produce_mocked()
    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(userspacegen.api.current_actor(), 'produce', mocked_produce)
    # The available RHSM repos
    monkeypatch.setattr(rhsm, 'get_available_repo_ids', lambda x: ['repoidA', 'repoidB', 'repoidC'])
    # The required RHEL repos based on the repo mapping and PES data + custom repos required by third party actors
    monkeypatch.setattr(userspacegen.api, 'consume', lambda x: iter([models.TargetRepositories(
        rhel_repos=[models.RHELTargetRepository(repoid='repoidX'),
                    models.RHELTargetRepository(repoid='repoidY')],
        custom_repos=[models.CustomTargetRepository(repoid='repoidCustom')])]))

    with pytest.raises(StopActorExecution):
        userspacegen.gather_target_repositories(None, indata)
    assert mocked_produce.called
    reports = [m.report for m in mocked_produce.model_instances if isinstance(m, reporting.Report)]
    inhibitors = [m for m in reports if 'inhibitor' in m.get('groups', ())]
    assert len(inhibitors) == 1
    assert inhibitors[0].get('title', '') == 'Cannot find required basic RHEL target repositories.'
    # Now test the case when either of AppStream and BaseOs is not available, upgrade should be inhibited
    mocked_produce = produce_mocked()
    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(userspacegen.api.current_actor(), 'produce', mocked_produce)
    monkeypatch.setattr(rhsm, 'get_available_repo_ids', lambda x: ['repoidA', 'repoidB', 'repoidC-appstream'])
    monkeypatch.setattr(userspacegen.api, 'consume', lambda x: iter([models.TargetRepositories(
        rhel_repos=[models.RHELTargetRepository(repoid='repoidC-appstream'),
                    models.RHELTargetRepository(repoid='repoidA')],
        custom_repos=[models.CustomTargetRepository(repoid='repoidCustom')])]))
    with pytest.raises(StopActorExecution):
        userspacegen.gather_target_repositories(None, indata)
    reports = [m.report for m in mocked_produce.model_instances if isinstance(m, reporting.Report)]
    inhibitors = [m for m in reports if 'inhibitor' in m.get('groups', ())]
    assert len(inhibitors) == 1
    assert inhibitors[0].get('title', '') == 'Cannot find required basic RHEL target repositories.'
    mocked_produce = produce_mocked()
    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(userspacegen.api.current_actor(), 'produce', mocked_produce)
    monkeypatch.setattr(rhsm, 'get_available_repo_ids', lambda x: ['repoidA', 'repoidB', 'repoidC-baseos'])
    monkeypatch.setattr(userspacegen.api, 'consume', lambda x: iter([models.TargetRepositories(
        rhel_repos=[models.RHELTargetRepository(repoid='repoidC-baseos'),
                    models.RHELTargetRepository(repoid='repoidA')],
        custom_repos=[models.CustomTargetRepository(repoid='repoidCustom')])]))
    with pytest.raises(StopActorExecution):
        userspacegen.gather_target_repositories(None, indata)
    reports = [m.report for m in mocked_produce.model_instances if isinstance(m, reporting.Report)]
    inhibitors = [m for m in reports if 'inhibitor' in m.get('groups', ())]
    assert len(inhibitors) == 1
    assert inhibitors[0].get('title', '') == 'Cannot find required basic RHEL target repositories.'


def mocked_consume_data():
    packages = {'dnf', 'dnf-command(config-manager)', 'pkgA', 'pkgB'}
    rhsm_info = _RHSMINFO_MSG
    rhui_info = _RHUIINFO_MSG
    xfs_info = models.XFSPresence()
    storage_info = models.StorageInfo()
    custom_repofiles = []
    files = set()
    fields = [
        'packages',
        'rhsm_info',
        'rhui_info',
        'xfs_info',
        'storage_info',
        'custom_repofiles',
        'files'
    ]

    return namedtuple('TestInData', fields)(
                packages,
                rhsm_info,
                rhui_info,
                xfs_info,
                storage_info,
                custom_repofiles,
                files,
    )


# TODO: come up with additional tests for the main function
def test_perform_ok(monkeypatch):
    repoids = ['repoidX', 'repoidY']
    monkeypatch.setattr(userspacegen, '_InputData', mocked_consume_data)
    monkeypatch.setattr(userspacegen, '_get_product_certificate_path', lambda: _DEFAULT_CERT_PATH)
    monkeypatch.setattr(overlaygen, 'create_source_overlay', MockedMountingBase)
    monkeypatch.setattr(userspacegen, '_gather_target_repositories', lambda *x: repoids)
    monkeypatch.setattr(userspacegen, '_create_target_userspace', lambda *x: None)
    monkeypatch.setattr(userspacegen.api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(userspacegen.api, 'produce', produce_mocked())
    monkeypatch.setattr(repofileutils, 'get_repodirs', lambda: ['/etc/yum.repos.d'])
    userspacegen.perform()
    msg_target_repos = models.UsedTargetRepositories(
        repos=[models.UsedTargetRepository(repoid=repo) for repo in repoids])
    assert userspacegen.api.produce.called == 3
    assert isinstance(userspacegen.api.produce.model_instances[0], models.TMPTargetRepositoriesFacts)
    assert userspacegen.api.produce.model_instances[1] == msg_target_repos
    # this one is full of constants, so it's safe to check just the instance
    assert isinstance(userspacegen.api.produce.model_instances[2], models.TargetUserSpaceInfo)



import os
import shutil
import tempfile

import distro
import pytest

from leapp.libraries.common import gpg
from leapp.libraries.common.testutils import CurrentActorMocked
from leapp.libraries.stdlib import api
from leapp.models import GpgKey, InstalledRPM, RPM


@pytest.mark.parametrize('target, product_type, exp', [
    ('8.6', 'beta', '../../files/rpm-gpg/8beta'),
    ('8.8', 'htb', '../../files/rpm-gpg/8'),
    ('9.0', 'beta', '../../files/rpm-gpg/9beta'),
    ('9.2', 'ga', '../../files/rpm-gpg/9'),
])
def test_get_path_to_gpg_certs(monkeypatch, target, product_type, exp):
    current_actor = CurrentActorMocked(dst_ver=target,
                                       envars={'LEAPP_DEVEL_TARGET_PRODUCT_TYPE': product_type})
    monkeypatch.setattr(api, 'current_actor', current_actor)

    p = gpg.get_path_to_gpg_certs()
    assert p == exp


def is_rhel7():
    return int(distro.major_version()) < 8


@pytest.mark.skipif(distro.id() not in ("rhel", "centos"), reason="Requires RHEL or CentOS for valid results.")
def test_gpg_show_keys(loaded_leapp_repository, monkeypatch):
    src = '7.9' if is_rhel7() else '8.6'
    current_actor = CurrentActorMocked(src_ver=src)
    monkeypatch.setattr(api, 'current_actor', current_actor)

    # python2 compatibility :/
    dirpath = tempfile.mkdtemp()

    # using GNUPGHOME env should avoid gnupg modifying the system
    os.environ['GNUPGHOME'] = dirpath

    try:
        # non-existing file
        non_existent_path = os.path.join(dirpath, 'nonexistent')
        res = gpg._gpg_show_keys(non_existent_path)
        if is_rhel7():
            err_msg = "gpg: can't open `{}'".format(non_existent_path)
        else:
            err_msg = "gpg: can't open '{}': No such file or directory\n".format(non_existent_path)
        assert not res['stdout']
        assert err_msg in res['stderr']
        assert res['exit_code'] == 2

        fp = gpg._parse_fp_from_gpg(res)
        assert fp == []

        # no gpg data found
        no_key_path = os.path.join(dirpath, "no_key")
        with open(no_key_path, "w") as f:
            f.write('test')

        res = gpg._gpg_show_keys(no_key_path)
        if is_rhel7():
            err_msg = ('gpg: no valid OpenPGP data found.\n'
                       'gpg: processing message failed: Unknown system error\n')
        else:
            err_msg = 'gpg: no valid OpenPGP data found.\n'
        assert not res['stdout']
        assert res['stderr'] == err_msg
        assert res['exit_code'] == 2

        fp = gpg._parse_fp_from_gpg(res)
        assert fp == []

        # with some test data now -- rhel9 release key
        # rhel9_key_path = os.path.join(api.get_common_folder_path('rpm-gpg'), '9')
        cur_dir = os.path.dirname(os.path.abspath(__file__))
        rhel9_key_path = os.path.join(cur_dir, '..', '..', 'files', 'rpm-gpg', '9',
                                      'RPM-GPG-KEY-redhat-release')
        res = gpg._gpg_show_keys(rhel9_key_path)
    finally:
        shutil.rmtree(dirpath)

    if is_rhel7():
        assert len(res['stdout']) == 4
        assert res['stdout'][0] == ('pub:-:4096:1:199E2F91FD431D51:1256212795:::-:'
                                    'Red Hat, Inc. (release key 2) <security@redhat.com>:')
        assert res['stdout'][1] == 'fpr:::::::::567E347AD0044ADE55BA8A5F199E2F91FD431D51:'
        assert res['stdout'][2] == ('pub:-:4096:1:5054E4A45A6340B3:1646863006:::-:'
                                    'Red Hat, Inc. (auxiliary key 3) <security@redhat.com>:')
        assert res['stdout'][3] == 'fpr:::::::::7E4624258C406535D56D6F135054E4A45A6340B3:'
    else:
        assert len(res['stdout']) == 6
        assert res['stdout'][0] == 'pub:-:4096:1:199E2F91FD431D51:1256212795:::-:::scSC::::::23::0:'
        assert res['stdout'][1] == 'fpr:::::::::567E347AD0044ADE55BA8A5F199E2F91FD431D51:'
        assert res['stdout'][2] == ('uid:-::::1256212795::DC1CAEC7997B3575101BB0FCAAC6191792660D8F::'
                                    'Red Hat, Inc. (release key 2) <security@redhat.com>::::::::::0:')
        assert res['stdout'][3] == 'pub:-:4096:1:5054E4A45A6340B3:1646863006:::-:::scSC::::::23::0:'
        assert res['stdout'][4] == 'fpr:::::::::7E4624258C406535D56D6F135054E4A45A6340B3:'
        assert res['stdout'][5] == ('uid:-::::1646863006::DA7F68E3872D6E7BDCE05225E7EB5F3ACDD9699F::'
                                    'Red Hat, Inc. (auxiliary key 3) <security@redhat.com>::::::::::0:')

    err = '{}/trustdb.gpg: trustdb created'.format(dirpath)
    assert err in res['stderr']
    assert res['exit_code'] == 0

    # now, parse the output too
    fp = gpg._parse_fp_from_gpg(res)
    assert fp == ['fd431d51', '5a6340b3']


@pytest.mark.parametrize('res, exp', [
    ({'exit_code': 2, 'stdout': '', 'stderr': ''}, []),
    ({'exit_code': 2, 'stdout': '', 'stderr': 'bash: gpg2: command not found...'}, []),
    ({'exit_code': 0, 'stdout': 'Some other output', 'stderr': ''}, []),
    ({'exit_code': 0, 'stdout': ['Some other output', 'other line'], 'stderr': ''}, []),
    ({'exit_code': 0, 'stdout': ['pub:-:4096:1:199E2F91FD431D:'], 'stderr': ''}, []),
    ({'exit_code': 0, 'stdout': ['pub:-:4096:1:5054E4A45A6340B3:1..'], 'stderr': ''}, ['5a6340b3']),
])
def test_parse_fp_from_gpg(res, exp):
    fp = gpg._parse_fp_from_gpg(res)
    assert fp == exp


def test_pubkeys_from_rpms():
    installed_rpms = InstalledRPM(
        items=[
            RPM(name='gpg-pubkey',
                version='9570ff31',
                release='5e3006fb',
                epoch='0',
                packager='Fedora (33) <fedora-33-primary@fedoraproject.org>',
                arch='noarch',
                pgpsig=''),
            RPM(name='rpm',
                version='4.17.1',
                release='3.fc35',
                epoch='0',
                packager='Fedora Project',
                arch='x86_64',
                pgpsig='RSA/SHA256, Tue 02 Aug 2022 03:12:43 PM CEST, Key ID db4639719867c58f'),
        ],
    )
    assert gpg.get_pubkeys_from_rpms(installed_rpms) == [GpgKey(fingerprint='9570ff31', rpmdb=True)]



from leapp.libraries.common.config import mock_configs
from leapp.models import FirmwareFacts, GrubInfo
from leapp.reporting import Report

NO_GRUB = 'Leapp could not identify where GRUB2 core is located'
GRUB = 'GRUB2 core will be automatically updated during the upgrade'


def test_actor_update_grub(current_actor_context):
    current_actor_context.feed(FirmwareFacts(firmware='bios'))
    current_actor_context.feed(GrubInfo(orig_devices=['/dev/vda', '/dev/vdb']))
    current_actor_context.run(config_model=mock_configs.CONFIG)
    assert current_actor_context.consume(Report)
    assert current_actor_context.consume(Report)[0].report['title'].startswith(GRUB)


def test_actor_no_grub_device(current_actor_context):
    current_actor_context.feed(FirmwareFacts(firmware='bios'))
    current_actor_context.feed(GrubInfo())
    current_actor_context.run(config_model=mock_configs.CONFIG)
    assert current_actor_context.consume(Report)
    assert current_actor_context.consume(Report)[0].report['title'].startswith(NO_GRUB)


def test_actor_with_efi(current_actor_context):
    current_actor_context.feed(FirmwareFacts(firmware='efi'))
    current_actor_context.run(config_model=mock_configs.CONFIG)
    assert not current_actor_context.consume(Report)


def test_s390x(current_actor_context):
    current_actor_context.feed(FirmwareFacts(firmware='bios'))
    current_actor_context.feed(GrubInfo(orig_devices=['/dev/vda', '/dev/vdb']))
    current_actor_context.run(config_model=mock_configs.CONFIG_S390X)
    assert not current_actor_context.consume(Report)



from leapp.actors import Actor
from leapp.libraries.actor import scanmemory
from leapp.models import MemoryInfo
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanMemory(Actor):
    """Scan Memory of the machine."""

    name = 'scanmemory'
    consumes = ()
    produces = (MemoryInfo,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        scanmemory.process()



from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common.config import architecture, version
from leapp.libraries.stdlib import api
from leapp.models import KernelCmdline, RoceDetected

FMT_LIST_SEPARATOR = '\n    - {}'
DOC_URL = 'https://red.ht/predictable-network-interface-device-names-on-the-system-z-platform'


def is_kernel_arg_set():
    """
    Return True if the system is booted with net.naming-scheme=rhel-8.7

    Important: it's really expected the argument is rhel-8.7 always.
    So not rhel-8.8, rhel-9.0, ... etc.
    """
    kernel_args = next(api.consume(KernelCmdline), None)
    if not kernel_args:
        # This is theoretical. If this happens, something is terribly wrong
        # already - so raising the hard error.
        raise StopActorExecutionError('Missing the KernelCmdline message!')
    for param in kernel_args.parameters:
        if param.key != 'net.naming-scheme':
            continue
        if param.value == 'rhel-8.7':
            return True
        api.current_logger().warning(
            'Detected net.naming-scheme with unexpected value: {}'
            .format(param.value)
        )
        return False
    return False


def _fmt_list(items):
    return ''.join([FMT_LIST_SEPARATOR.format(i) for i in items])


def _report_old_version(roce):
    roce_nics = roce.roce_nics_connected + roce.roce_nics_connecting
    reporting.create_report([
        reporting.Title('A newer version of RHEL 8 is required for the upgrade with RoCE.'),
        reporting.Summary(
            'The RHEL 9 system uses different network schemes for NIC names'
            ' than RHEL 8.'
            ' RHEL {version} does not provide functionality to be able'
            ' to set the system configuration in a way the network interface'
            ' names used by RoCE are persistent on both (RHEL 8 and RHEL 9)'
            ' systems.'
            ' The in-place upgrade from the current version of RHEL to RHEL 9'
            ' will break the RoCE network configuration.'
            '\n\nRoCE detected on following NICs:{nics}'
            .format(
                version=version.get_source_version(),
                nics=_fmt_list(roce_nics)
            )
        ),
        reporting.Remediation(hint=(
            'Update the system to RHEL 8.8 or newer using DNF and then reboot'
            ' the system prior the in-place upgrade to RHEL 9.'
        )),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([
            reporting.Groups.INHIBITOR,
            reporting.Groups.ACCESSIBILITY,
            reporting.Groups.SANITY,
        ]),
    ])


def _report_wrong_setup(roce):
    roce_nics = roce.roce_nics_connected + roce.roce_nics_connecting
    reporting.create_report([
        reporting.Title('Invalid RoCE configuration for the in-place upgrade'),
        reporting.Summary(
            'The RHEL 9 system uses different network schemes for NIC names'
            ' than RHEL 8.'
            ' The below listed RoCE NICs need to be reconfigured to the new'
            ' interface naming scheme in order to prevent loss of network'
            ' access to your system via these interfaces after the upgrade.'
            ' For more information, see: {url}'
            '\n\nRoCE detected on the following NICs:{nics}'
            .format(nics=_fmt_list(roce_nics), url=DOC_URL)
        ),
        reporting.Remediation(hint=(
            'Prerequisite for upgrading to RHEL9.x:'
            'In RHEL 8, all RoCE cards must be configured with the interface'
            ' names they should have in RHEL9.x.\n'
            'For more information, see chapter 1.4 of the RHEL8 Product'
            ' Documentation (see the attached link) and follow these steps:\n'
            '1.) determine the current interface device names of the RoCE'
            ' cards that are in "connected to" or in "connecting" state\n'
            '2.) determine if UID uniqueness is set for these cards\n'
            '3.) compute new interface device names from the UID or the'
            ' function ID, respectively\n'
            '4.) change the network interface device names in ifcfg'
            ' files\n'
            '5.) set the kernel parameter net.naming-scheme=rhel-8.7 in the'
            ' effective .conf file in /boot/loader/entries\n'
            '6.) adjust other settings that rely on the interface device names'
            ' (e.g. firewall) by changing the interface device names'
            ' accordingly\n'
            '7.) run `zipl -V` and reboot the system\n'
            '8.) check your network connectivity\n'
            '\n'
            'Caution: Creating an incorrect configuration might cause the loss'
            ' of your network connection after reboot!'
        )),
        reporting.ExternalLink(
            title='Predictable network interface device names on the System z platform',
            url=DOC_URL),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([
            reporting.Groups.INHIBITOR,
            reporting.Groups.ACCESSIBILITY,
            reporting.Groups.SANITY,
        ]),
    ])


def process():
    if not architecture.matches_architecture(architecture.ARCH_S390X):
        # The check is valid only on S390X architecture
        return
    roce = next(api.consume(RoceDetected), None)
    if not roce or not (roce.roce_nics_connected or roce.roce_nics_connecting):
        # No used RoCE detected - nothing to do
        api.current_logger().debug('Skipping RoCE checks: No RoCE card detected.')
        return
    if version.matches_source_version('<= 8.6'):
        _report_old_version(roce)
    if not is_kernel_arg_set():
        _report_wrong_setup(roce)



import datetime
import json
import os
import sys
from functools import wraps
from itertools import chain

from leapp import FULL_VERSION
from leapp.libraries.stdlib.call import _call
from leapp.utils.audit import get_messages

try:
    from json.decoder import JSONDecodeError  # pylint: disable=ungrouped-imports
except ImportError:
    JSONDecodeError = ValueError


def runs_in_container():
    """
    Check if the current process is running inside a container

    :return: True if the process is running inside a container, False otherwise
    """
    return os.path.exists('/run/host/container-manager')


def _flattened(d):
    """ Flatten nested dicts and lists into a single dict """
    def expand(key, value):
        if isinstance(value, dict):
            return [(key + '.' + k, v) for k, v in _flattened(value).items()]
        if isinstance(value, list):
            return chain(*[expand(key + '.' + str(i), v) for i, v in enumerate(value)])
        return [(key, value)]
    items = [item for k, v in d.items() for item in expand(k, v)]
    return dict(items)


class _BreadCrumbs(object):
    def __init__(self, activity):
        self._crumbs = {
            'activity': activity,
            'packages': self._get_packages(),
            'leapp_file_changes': [],
            'executed': ' '.join([v if ' ' not in v else '"{}"'.format(v) for v in sys.argv]),
            'success': True,
            'activity_started': datetime.datetime.utcnow().isoformat() + 'Z',
            'activity_ended': datetime.datetime.utcnow().isoformat() + 'Z',
            'source_os': '',
            'target_os': '',
            'env': dict(),
            'run_id': '',
            'version': FULL_VERSION,
        }

    def fail(self):
        self._crumbs['success'] = False

    def _save_rhsm_facts(self, activities):
        if not os.path.isdir('/etc/rhsm/facts'):
            if not os.path.exists('/etc/rhsm'):
                # If there's no /etc/rhsm folder just skip it
                return
            try:
                os.mkdir('/etc/rhsm/facts')
            except OSError as e:
                if e.errno == 17:
                    # The directory already exists which is all we need.
                    pass
        try:
            with open('/etc/rhsm/facts/leapp.facts', 'w') as f:
                json.dump(_flattened({
                    'leapp': [
                        activity for activity in activities
                        if activity.get('activity', '') in ('preupgrade', 'upgrade')]
                    }), f, indent=4)
            self._commit_rhsm_facts()
        except OSError:
            # We don't care about failing to 'create' the file here
            # even though it shouldn't though, just ignore it
            pass

    def _commit_rhsm_facts(self):
        if runs_in_container():
            return
        cmd = ['/usr/sbin/subscription-manager', 'facts', '--update']
        try:
            _call(cmd, lambda x, y: None, lambda x, y: None)
        except (OSError, ValueError, TypeError):
            # We don't care about errors here, just ignore them
            pass

    def save(self):
        self._crumbs['run_id'] = os.environ.get('LEAPP_EXECUTION_ID', 'N/A')
        self._crumbs['leapp_file_changes'].extend(self._verify_leapp_pkgs())
        messages = get_messages(('IPUConfig',), self._crumbs['run_id'])
        versions = json.loads((messages or [{}])[0].get('message', {}).get(
            'data', '{}')).get('version', {'target': 'N/A', 'source': 'N/A'})
        self._crumbs['target_os'] = 'Red Hat Enterprise Linux {target}'.format(**versions)
        self._crumbs['source_os'] = 'Red Hat Enterprise Linux {source}'.format(**versions)
        self._crumbs['activity_ended'] = datetime.datetime.utcnow().isoformat() + 'Z'
        self._crumbs['env'] = {k: v for k, v in os.environ.items() if k.startswith('LEAPP_')}
        try:
            with open('/etc/migration-results', 'a+') as crumbs:
                crumbs.seek(0)
                doc = {'activities': []}
                try:
                    content = json.load(crumbs)
                    if isinstance(content, dict):
                        if isinstance(content.get('activities', None), list):
                            doc = content
                except JSONDecodeError:
                    # Expected to happen when /etc/migration-results is still empty or does not yet exist
                    pass
                doc['activities'].append(self._crumbs)
                crumbs.seek(0)
                crumbs.truncate()
                json.dump(doc, crumbs, indent=2, sort_keys=True)
                crumbs.write('\n')
                if os.environ.get('LEAPP_NO_RHSM_FACTS', '0') != '1':
                    self._save_rhsm_facts(doc['activities'])
        except OSError:
            sys.stderr.write('WARNING: Could not write to /etc/migration-results\n')

    def _get_packages(self):
        cmd = ['/bin/bash', '-c', 'rpm -qa --queryformat="%{nevra} %{SIGPGP:pgpsig}\n" | grep -Ee "leapp|snactor"']
        res = _call(cmd, lambda x, y: None, lambda x, y: None)
        if res.get('exit_code', None) == 0:
            if res.get('stdout', None):
                return [{'nevra': t[0], 'signature': t[1]}
                        for t in [line.strip().split(' ', 1) for line in res['stdout'].split('\n') if line.strip()]]
        return []

    def _verify_leapp_pkgs(self):
        if not os.environ.get('LEAPP_IPU_IN_PROGRESS'):
            return []
        upg_path = os.environ.get('LEAPP_IPU_IN_PROGRESS').split('to')
        cmd = ['/bin/bash', '-c', 'rpm -V leapp leapp-upgrade-el{}toel{}'.format(upg_path[0], upg_path[1])]
        res = _call(cmd, lambda x, y: None, lambda x, y: None)
        if res.get('exit_code', None) == 1:
            if res.get('stdout', None):
                return [{'result': t[0], 'file_name': t[1]}
                        for t in [line.strip().split(' ', 1) for line in res['stdout'].split('\n') if line.strip()]]
        return []


def produces_breadcrumbs(f):
    """
    Ensures that `/etc/migration-results` gets produced on every invocation of `leapp upgrade` & `leapp preupgrade`

    Every execution of the upgrade will have their own entry in the /etc/migration-results file.
    For a user flow like: leapp preupgrade && leapp upgrade && reboot there should be 5 new entries in the file:

    1. leapp preupgrade
    2. leapp upgrade (Source OS)
    3. leapp upgrade (Initram Phase - Until including RPM transaction)
    4. leapp upgrade (Initram Phase - Post RPM Transaction)
    5. leapp upgrade (Target OS - First Boot)

    Depending on future design changes of the IPU Worklow, the output may vary.
    """
    @wraps(f)
    def wrapper(*args, **kwargs):
        breadcrumbs = _BreadCrumbs(activity=f.__name__)
        try:
            return f(*args, breadcrumbs=breadcrumbs, **kwargs)
        except SystemExit as e:
            if e.code != 0:
                breadcrumbs.fail()
            raise
        except BaseException:
            breadcrumbs.fail()
            raise
        finally:
            breadcrumbs.save()
    return wrapper



import pytest

from leapp.libraries.actor import scansystemdtarget
from leapp.libraries.common import systemd
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api, CalledProcessError
from leapp.models import (
    SystemdBrokenSymlinksTarget,
    SystemdServiceFile,
    SystemdServicePreset,
    SystemdServicesInfoTarget,
    SystemdServicesPresetInfoTarget
)

_BROKEN_SYMLINKS = [
    "/etc/systemd/system/multi-user.target.wants/vdo.service",
    "/etc/systemd/system/multi-user.target.wants/rngd.service"
]

_SERVICE_FILES = [
    SystemdServiceFile(name='getty@.service', state='enabled'),
    SystemdServiceFile(name='vdo.service', state='disabled')
]

_PRESETS = [
    SystemdServicePreset(service='getty@.service', state='enable'),
    SystemdServicePreset(service='vdo.service', state='disable'),
]


@pytest.mark.parametrize(
    ('broken_symlinks', 'files', 'presets'),
    (
        (_BROKEN_SYMLINKS, _SERVICE_FILES, _PRESETS),
        ([], [], [])
    )
)
def test_message_produced(monkeypatch, broken_symlinks, files, presets):

    def scan_broken_symlinks_mocked():
        return broken_symlinks

    def get_service_files_mocked():
        return files

    def get_system_service_preset_files_mocked(service_files, ignore_invalid_entries):
        return presets

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(systemd, 'get_broken_symlinks', scan_broken_symlinks_mocked)
    monkeypatch.setattr(systemd, 'get_service_files', get_service_files_mocked)
    monkeypatch.setattr(systemd, 'get_system_service_preset_files', get_system_service_preset_files_mocked)

    scansystemdtarget.scan()

    assert api.produce.called
    assert api.produce.model_instances[0].broken_symlinks == broken_symlinks
    assert api.produce.model_instances[1].service_files == files
    assert api.produce.model_instances[2].presets == presets


_CALL_PROC_ERR = CalledProcessError(
    message='BooCalled',
    command=['find'],
    result={
        'stdout': 'stdout',
        'stderr': 'stderr',
        'exit_code': 1,
        'signal': 1,
        'pid': 1,
    }
)


class GetOrRaise(object):
    def __init__(self, value):
        self.value = value

    def __call__(self, *dummyArgs, **dummy):
        if isinstance(self.value, list):
            return self.value
        raise self.value


@pytest.mark.parametrize('symlinks', [OSError('Boo'), _CALL_PROC_ERR, []])
@pytest.mark.parametrize('files', [_CALL_PROC_ERR, []])
@pytest.mark.parametrize('presets', [OSError('Boo'), _CALL_PROC_ERR, []])
def test_exception_handling(monkeypatch, symlinks, files, presets):

    def check_msg(input_data, msg_type, msgs, is_msg_expected):
        for msg in msgs.model_instances:
            if isinstance(msg, msg_type):
                return is_msg_expected
        return not is_msg_expected

    if symlinks == files == presets == []:
        # covered by test above
        return

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(systemd, 'get_broken_symlinks', GetOrRaise(symlinks))
    monkeypatch.setattr(systemd, 'get_service_files', GetOrRaise(files))
    monkeypatch.setattr(systemd, 'get_system_service_preset_files', GetOrRaise(presets))
    scansystemdtarget.scan()
    assert check_msg(symlinks, SystemdBrokenSymlinksTarget, api.produce, isinstance(symlinks, list))
    assert check_msg(files, SystemdServicesInfoTarget, api.produce, isinstance(files, list))
    is_msg_expected = isinstance(files, list) and isinstance(presets, list)
    assert check_msg(presets, SystemdServicesPresetInfoTarget, api.produce, is_msg_expected)



import os.path

from leapp.libraries.stdlib import api
from leapp.models import DistributionSignedRPM, RpmTransactionTasks


def load_tasks_file(path, logger):
    # Loads the given file and converts it to a deduplicated list of strings that are stripped
    if os.path.isfile(path):
        try:
            with open(path, 'r') as f:
                return list(
                    {entry.strip() for entry in f.read().split('\n') if entry.strip() and
                        not entry.strip().startswith('#')}
                )
        except IOError as e:
            logger.warning('Failed to open %s to load additional transaction data. Error: %s', path, str(e))
    return []


def load_tasks(base_dir, logger):
    # Loads configuration files to_install, to_keep, and to_remove from the given base directory
    rpms = next(api.consume(DistributionSignedRPM))
    rpm_names = [rpm.name for rpm in rpms.items]
    to_install = load_tasks_file(os.path.join(base_dir, 'to_install'), logger)
    # we do not want to put into rpm transaction what is already installed (it will go to "to_upgrade" bucket)
    to_install_filtered = [pkg for pkg in to_install if pkg not in rpm_names]

    filtered = set(to_install) - set(to_install_filtered)
    if filtered:
        api.current_logger().debug(
            'The following packages from "to_install" file will be ignored as they are already installed:'
            '\n- ' + '\n- '.join(filtered))

    return RpmTransactionTasks(
        to_install=to_install_filtered,
        to_keep=load_tasks_file(os.path.join(base_dir, 'to_keep'), logger),
        to_remove=load_tasks_file(os.path.join(base_dir, 'to_remove'), logger))



from leapp import reporting
from leapp.libraries.common import rhsm
from leapp.libraries.stdlib import api


def _report_set_release():
    target_version = api.current_actor().configuration.version.target
    reporting.create_report([
        reporting.Title(
            'The subscription-manager release is going to be set after the upgrade'),
        reporting.Summary(
            'After the upgrade has completed the release of the subscription-manager will be set to {release}.'
            ' This will ensure that you will receive and keep the version you choose to upgrade to.'
            .format(release=target_version)
        ),
        reporting.Severity(reporting.Severity.LOW),
        reporting.Remediation(
            hint='If you wish to receive updates for the latest released version of the target system, '
                 'run `subscription-manager release --unset` after the upgrade.'
        ),
        reporting.Groups([reporting.Groups.UPGRADE_PROCESS]),
        reporting.RelatedResource('package', 'subscription-manager')
    ])


def _report_unhandled_release():
    # TODO: set the POST group after it's created.
    target_version = api.current_actor().configuration.version.target
    hint_command = 'subscription-manager release --set {}'.format(target_version)
    # FIXME: This should use Dialogs and Answers to offer post-upgrade remediation
    # so that users can choose whether to --set or --unset the release number
    hint = 'Set the new release (or unset it) after the upgrade using subscription-manager: ' + hint_command
    reporting.create_report([
        reporting.Title(
            'The subscription-manager release is going to be kept as it is during the upgrade'),
        reporting.Summary(
            'The upgrade is executed with the --no-rhsm option (or with'
            ' the LEAPP_NO_RHSM environment variable). In this case, the subscription-manager'
            ' will not be configured during the upgrade. If the system is subscribed and release'
            ' is set already, you could encounter issues to get RHEL content using DNF/YUM'
            ' after the upgrade.'
        ),
        reporting.Severity(reporting.Severity.LOW),
        reporting.Remediation(hint=hint),
        reporting.Groups([reporting.Groups.UPGRADE_PROCESS]),
        reporting.RelatedResource('package', 'subscription-manager')
    ])


def process():
    if rhsm.skip_rhsm():
        _report_unhandled_release()
    else:
        _report_set_release()



from leapp.actors import Actor
from leapp.libraries.actor import upgradeinitramfsgenerator
from leapp.models import RequiredUpgradeInitramPackages  # deprecated
from leapp.models import UpgradeDracutModule  # deprecated
from leapp.models import (
    BootContent,
    FIPSInfo,
    TargetOSInstallationImage,
    TargetUserSpaceInfo,
    TargetUserSpaceUpgradeTasks,
    UpgradeInitramfsTasks,
    UsedTargetRepositories
)
from leapp.tags import InterimPreparationPhaseTag, IPUWorkflowTag


class UpgradeInitramfsGenerator(Actor):
    """
    Creates the upgrade initramfs

    Creates an initram disk within a systemd-nspawn container using the target
    system userspace, including new kernel. The creation of the initram disk
    can be influenced with the UpgradeInitramfsTasks message (e.g. specifying
    what files or dracut modules should be installed in the upgrade initramfs)

    See the UpgradeInitramfsTasks model for more details.
    """

    name = 'upgrade_initramfs_generator'
    consumes = (
        FIPSInfo,
        RequiredUpgradeInitramPackages,  # deprecated
        TargetOSInstallationImage,
        TargetUserSpaceInfo,
        TargetUserSpaceUpgradeTasks,
        UpgradeDracutModule,  # deprecated
        UpgradeInitramfsTasks,
        UsedTargetRepositories,
    )
    produces = (BootContent,)
    tags = (IPUWorkflowTag, InterimPreparationPhaseTag)

    def process(self):
        upgradeinitramfsgenerator.process()



from leapp.actors import Actor
from leapp.libraries.actor.checkkdeapps import get_kde_apps_info
from leapp.models import InstalledKdeAppsFacts, InstalledRPM
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class CheckKdeApps(Actor):
    """
    Actor checks which KDE apps are installed.
    """

    name = 'check_kde_apps'
    consumes = (InstalledRPM,)
    produces = (InstalledKdeAppsFacts,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        app_facts = get_kde_apps_info()
        self.produce(InstalledKdeAppsFacts(
            installed_apps=app_facts))



from leapp import reporting
from leapp.libraries.common import config, rhsm
from leapp.libraries.common.config.version import get_target_major_version
from leapp.libraries.stdlib import api
from leapp.models import CustomTargetRepositoryFile, RHUIInfo, TargetRepositories

# TODO: we need to provide this path in a shared library
CUSTOM_REPO_PATH = '/etc/leapp/files/leapp_upgrade_repositories.repo'


def _any_custom_repo_defined():
    for tr in api.consume(TargetRepositories):
        if tr.custom_repos:
            return True
    return False


def _the_custom_repofile_defined():
    for ctrf in api.consume(CustomTargetRepositoryFile):
        if ctrf and ctrf.file == CUSTOM_REPO_PATH:
            return True
    return False


def _the_enablerepo_option_used():
    return config.get_env('LEAPP_ENABLE_REPOS', None) is not None


def process():
    target_major_version = get_target_major_version()

    if target_major_version == '8':
        ipu_doc_url = 'https://red.ht/upgrading-rhel7-to-rhel8-main-official-doc'
    elif target_major_version == '9':
        ipu_doc_url = 'https://red.ht/upgrading-rhel8-to-rhel9-main-official-doc'

    rhui_info = next(api.consume(RHUIInfo), None)

    if not rhsm.skip_rhsm() or rhui_info:
        # getting RH repositories through RHSM or RHUI; resolved by seatbelts
        # implemented in other actors
        return

    # rhsm skipped; take your seatbelts please
    is_ctr = _any_custom_repo_defined()
    is_ctrf = _the_custom_repofile_defined()
    is_re = _the_enablerepo_option_used()
    if not is_ctr:
        # no rhsm, no custom repositories.. this will really not work :)
        # TODO: add link to the RH article about use of custom repositories!!
        # NOTE: we can put here now the link to the main document, as this
        # will be described there or at least the link to the right document
        # will be delivered here.
        if is_ctrf:
            summary_ctrf = '\n\nThe custom repository file has been detected. Maybe it is empty?'
        else:
            summary_ctrf = ''
        reporting.create_report([
            reporting.Title('Using RHSM has been skipped but no custom or RHUI repositories have been delivered.'),
            reporting.Summary(
                'Leapp is run in the mode when the Red Hat Subscription Manager'
                ' is not used (the --no-rhsm option or the LEAPP_NO_RHSM=1'
                ' environment variable has been set) so leapp is not able to'
                ' obtain YUM/DNF repositories with the content for the target'
                ' system in the standard way. The content has to be delivered'
                ' either by user manually or, in case of public clouds, by a'
                ' special Leapp package for RHUI environments.'
                ),
            reporting.Remediation(hint=(
                'Create the repository file according to instructions in the'
                ' referred document on the following path with all'
                ' repositories that should be used during the upgrade: "{}".'
                '\n\n{}'
                .format(CUSTOM_REPO_PATH, summary_ctrf)
            )),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.SANITY]),
            reporting.Groups([reporting.Groups.INHIBITOR]),
            reporting.ExternalLink(url=ipu_doc_url, title='UPGRADING TO RHEL {}'.format(target_major_version)),
            reporting.RelatedResource('file', CUSTOM_REPO_PATH),
        ])
    elif not (is_ctrf or is_re):
        # Some custom repositories have been discovered, but the custom repo
        # file not - neither the --enablerepo option is used. Inform about
        # the official recommended way.
        reporting.create_report([
            reporting.Title('Detected "CustomTargetRepositories" without using new provided mechanisms used.'),
            reporting.Summary(
                'Red Hat now provides an official way for using custom'
                ' repositories during the in-place upgrade through'
                ' the referred custom repository file or through the'
                ' --enablerepo option for leapp. The CustomTargetRepositories'
                ' have been produced from custom (own) actors?'
            ),
            reporting.Remediation(hint=(
                'Follow the new simple way to enable custom repositories'
                ' during the upgrade (see the referred document) or create'
                ' the empty custom repository file to acknowledge this report'
                ' message.'
            )),
            reporting.Severity(reporting.Severity.INFO),
            reporting.ExternalLink(url=ipu_doc_url, title='UPGRADING TO RHEL {}'.format(target_major_version)),
            reporting.RelatedResource('file', CUSTOM_REPO_PATH),
        ])



from leapp.models import GrubConfigError, Report
from leapp.utils import report

grub_cmdline_syntax_error = GrubConfigError(error_type=GrubConfigError.ERROR_GRUB_CMDLINE_LINUX_SYNTAX,
                                            files=['/etc/default/grub.cfg'])
grub_cmdline_syntax_error2 = GrubConfigError(error_type=GrubConfigError.ERROR_GRUB_CMDLINE_LINUX_SYNTAX,
                                             files=['/boot/grub2/grub.cfg', '/etc/default/someothergrub.cfg'])

grub_missing_newline_error = GrubConfigError(error_type=GrubConfigError.ERROR_MISSING_NEWLINE,
                                             files=['/etc/default/someothergrub.cfg'])
grub_missing_newline_error2 = GrubConfigError(error_type=GrubConfigError.ERROR_MISSING_NEWLINE,
                                              files=['/etc/default/grub'])

grub_corrupted_config = GrubConfigError(error_type=GrubConfigError.ERROR_CORRUPTED_GRUBENV,
                                        files=['/boot/grub2/grub.cfg', '/boot/efi/EFI/redhat/grub.cfg'])
grub_corrupted_config2 = GrubConfigError(error_type=GrubConfigError.ERROR_CORRUPTED_GRUBENV,
                                         files=['/boot/grub2/grub.cfg'])


def test_cmdline_syntax_error(current_actor_context):
    # Make sure that just 1 low priority report message is created with config files present.
    current_actor_context.feed(grub_cmdline_syntax_error)
    current_actor_context.feed(grub_cmdline_syntax_error2)
    current_actor_context.run()
    messages = current_actor_context.consume(Report)
    assert len(messages) == 1
    message = messages[0]
    assert 'Syntax error detected in grub configuration' in message.report['title']
    assert message.report['severity'] == 'low'
    assert message.report['detail']['related_resources'][0]['title'] == '/etc/default/grub.cfg'


def test_missing_newline(current_actor_context):
    # Make sure that just 1 low priority report message is created with config files present
    current_actor_context.feed(grub_missing_newline_error)
    current_actor_context.feed(grub_missing_newline_error2)
    current_actor_context.run()
    messages = current_actor_context.consume(Report)
    assert len(messages) == 1
    message = messages[0]
    assert 'Detected a missing newline at the end of grub configuration file' in message.report['title']
    assert message.report['severity'] == 'low'
    assert message.report['detail']['related_resources'][0]['title'] == '/etc/default/someothergrub.cfg'


def test_corrupted_config(current_actor_context):
    # Make sure that just 1 high priority report message is created with config files present
    current_actor_context.feed(grub_corrupted_config)
    current_actor_context.feed(grub_corrupted_config2)
    current_actor_context.run()
    messages = current_actor_context.consume(Report)
    assert len(messages) == 1
    message = messages[0]
    assert 'Detected a corrupted grubenv file' in message.report['title']
    assert message.report['severity'] == 'high'
    assert message.report['detail']['related_resources'][0]['title'] == '/boot/grub2/grub.cfg'
    assert message.report['detail']['related_resources'][1]['title'] == '/boot/efi/EFI/redhat/grub.cfg'
    assert report.is_inhibitor(message.report)



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class FirmwareFacts(Model):
    topic = SystemFactsTopic

    firmware = fields.StringEnum(['bios', 'efi'])
    """ System firmware interface (BIOS or EFI) """
    ppc64le_opal = fields.Nullable(fields.Boolean())
    """ Check OPAL presence to identify ppc64le bare metal systems """



from leapp.actors import Actor
from leapp.libraries.actor import applycustomdnfconf
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class ApplyCustomDNFConf(Actor):
    """
    Move /etc/leapp/files/dnf.conf to /etc/dnf/dnf.conf if it exists

    An actor in FactsPhase copies this file to the target userspace if present.
    In such case we also want to use the file on the target system.
    """
    name = "apply_custom_dnf_conf"
    consumes = ()
    produces = ()
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        applycustomdnfconf.process()



from leapp.actors import Actor
from leapp.models import TargetInitramfsTasks
from leapp.tags import IPUWorkflowTag, TargetTransactionChecksPhaseTag


class TMPActorToSatisfySanityChecks(Actor):
    """
    The actor does NOTHING but satisfy static sanity checks

    The actor is supposed to be removed in future once we resolve issue with
    sanity tests. See https://github.com/oamg/leapp/pull/680 for more details
    about the problems.
    """

    name = 'tmp_actor_to_satisfy_sanity_checks'
    consumes = ()
    produces = (TargetInitramfsTasks,)
    tags = (TargetTransactionChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        return



import json
import os

from leapp.libraries.common import mounting, utils
from leapp.libraries.stdlib import api
from leapp.models import fields, RepositoryData, RepositoryFile

try:
    import dnf
except ImportError:
    api.current_logger().warning('repofileutils.py: failed to import dnf')


def _parse_repository(repoid, repo_data):
    def asbool(x):
        return x == '1'
    prepared = {'repoid': repoid, 'additional_fields': {}}
    for key in repo_data.keys():
        if key in RepositoryData.fields:
            if isinstance(RepositoryData.fields[key], fields.Boolean):
                repo_data[key] = asbool(repo_data[key])
            prepared[key] = repo_data[key]
        else:
            prepared['additional_fields'][key] = repo_data[key]
    prepared['additional_fields'] = json.dumps(prepared['additional_fields'])
    return RepositoryData(**prepared)


def parse_repofile(repofile):
    """
    Parse the given repo file.

    :param repofile: Path to the repo file
    :type repofile: str
    :rtype: RepositoryFile
    """
    data = []
    with open(repofile, mode='r') as fp:
        cp = utils.parse_config(fp, strict=False)
        for repoid in cp.sections():
            data.append(_parse_repository(repoid, dict(cp.items(repoid))))
    return RepositoryFile(file=repofile, data=data)


def get_repodirs():
    """
    Return all directories yum scans for repository files, if they exist.
    By default, the possible paths on RHEL should be:
    ['/etc/yum.repos.d', '/etc/yum/repos.d', '/etc/distro.repos.d']

    ATTENTION: Requires the dnf module to be present.
    TODO: Get repodirs inside given context.
    """
    with dnf.base.Base() as base:
        base.conf.read(priority=dnf.conf.PRIO_MAINCONFIG)
        return list({os.path.realpath(d) for d in base.conf.reposdir if os.path.isdir(d)})


def get_parsed_repofiles(context=mounting.NotIsolatedActions(base_dir='/')):
    """
    Scan all repositories on the system.

    Repositories are scanned under repository directories (as reported by dnf)
    of the given context. By default the context is the host system.

    ATTENTION: Do not forget to ensure the redhat.repo file is regenerated
    by RHSM when used.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :rtype: List(RepositoryFile)
    """
    repofiles = []
    cmd = ['find', '-L'] + get_repodirs() + ['-maxdepth', '1', '-type', 'f', '-name', '*.repo']
    repofiles_paths = context.call(cmd, split=True)['stdout']
    for repofile_path in repofiles_paths:
        repofile = parse_repofile(context.full_path(repofile_path))
        # we want full path in cotext, not the real full path
        repofile.file = repofile_path
        repofiles.append(repofile)
    return repofiles


def _invert_dict(data):
    """{a: [b]} -> {b: [a]}"""
    inv_dict = {}
    for key in data.keys():
        for value in data[key]:
            inv_dict[value] = inv_dict.get(value, []) + [key]
    return inv_dict


def get_duplicate_repositories(repofiles):
    """
    Return dict of duplicate repositories {repoid: [repofile_path]}

    A repository is defined multiple times if it exists in multiple repofiles.
    Redefinition inside one repository file is ignored (same in DNF).

    :param repofiles:
    :type repofiles: List(RepositoryFile)
    :rtype: dict {repoid: repofilepath}
    """
    rf_repos = {repofile.file: [repo.repoid for repo in repofile.data] for repofile in repofiles}
    repos = _invert_dict(rf_repos)
    return {repo: set(rfiles) for repo, rfiles in repos.items() if len(set(rfiles)) > 1}



from leapp.models import DaemonList, TcpWrappersFacts, VsftpdConfig, VsftpdFacts
from leapp.reporting import Report
from leapp.snactor.fixture import current_actor_context
from leapp.utils.report import is_inhibitor


def test_actor_with_unsupported_tcpwrap_and_vsftpd_config(current_actor_context):
    config1 = VsftpdConfig(path='/etc/vsftpd/foo.conf', tcp_wrappers=False)
    config2 = VsftpdConfig(path='/etc/vsftpd/bar.conf', tcp_wrappers=True)
    vsftpd_facts = VsftpdFacts(configs=[config1, config2])
    daemon_list = DaemonList(value=['vsftpd'])
    tcpwrap_facts = TcpWrappersFacts(daemon_lists=[daemon_list])

    current_actor_context.feed(vsftpd_facts)
    current_actor_context.feed(tcpwrap_facts)
    current_actor_context.run()
    report_fields = current_actor_context.consume(Report)[0].report

    assert is_inhibitor(report_fields)
    assert 'foo.conf' not in report_fields['summary']
    assert 'bar.conf' in report_fields['summary']


def test_actor_with_unsupported_tcpwrap_multiple_unsupported_vsftpd_configs(current_actor_context):
    config1 = VsftpdConfig(path='/etc/vsftpd/foo.conf', tcp_wrappers=True)
    config2 = VsftpdConfig(path='/etc/vsftpd/bar.conf', tcp_wrappers=False)
    config3 = VsftpdConfig(path='/etc/vsftpd/goo.conf', tcp_wrappers=True)
    vsftpd_facts = VsftpdFacts(configs=[config1, config2, config3])
    daemon_list = DaemonList(value=['vsftpd'])
    tcpwrap_facts = TcpWrappersFacts(daemon_lists=[daemon_list])

    current_actor_context.feed(vsftpd_facts)
    current_actor_context.feed(tcpwrap_facts)
    current_actor_context.run()
    report_fields = current_actor_context.consume(Report)[0].report

    assert is_inhibitor(report_fields)
    assert 'foo.conf' in report_fields['summary']
    assert 'bar.conf' not in report_fields['summary']
    assert 'goo.conf' in report_fields['summary']


def test_actor_with_unsupported_tcpwrap_config(current_actor_context):
    config1 = VsftpdConfig(path='/etc/vsftpd/foo.conf', tcp_wrappers=False)
    config2 = VsftpdConfig(path='/etc/vsftpd/bar.conf', tcp_wrappers=None)
    vsftpd_facts = VsftpdFacts(configs=[config1, config2])
    daemon_list = DaemonList(value=['vsftpd'])
    tcpwrap_facts = TcpWrappersFacts(daemon_lists=[daemon_list])

    current_actor_context.feed(vsftpd_facts)
    current_actor_context.feed(tcpwrap_facts)
    current_actor_context.run()

    assert not current_actor_context.consume(Report)


def test_actor_with_unsupported_vsftpd_config(current_actor_context):
    config1 = VsftpdConfig(path='/etc/vsftpd/foo.conf', tcp_wrappers=False)
    config2 = VsftpdConfig(path='/etc/vsftpd/bar.conf', tcp_wrappers=True)
    vsftpd_facts = VsftpdFacts(configs=[config1, config2])
    daemon_list = DaemonList(value=['all', 'except', 'vsftpd'])
    tcpwrap_facts = TcpWrappersFacts(daemon_lists=[daemon_list])

    current_actor_context.feed(vsftpd_facts)
    current_actor_context.feed(tcpwrap_facts)
    current_actor_context.run()

    assert not current_actor_context.consume(Report)


def test_actor_with_supported_tcpwrap_and_vsftpd_config(current_actor_context):
    config1 = VsftpdConfig(path='/etc/vsftpd/foo.conf', tcp_wrappers=False)
    config2 = VsftpdConfig(path='/etc/vsftpd/bar.conf', tcp_wrappers=False)
    vsftpd_facts = VsftpdFacts(configs=[config1, config2])
    daemon_list = DaemonList(value=['all', 'except', 'vsftpd'])
    tcpwrap_facts = TcpWrappersFacts(daemon_lists=[daemon_list])

    current_actor_context.feed(vsftpd_facts)
    current_actor_context.feed(tcpwrap_facts)
    current_actor_context.run()

    assert not current_actor_context.consume(Report)



import grp
import pwd

import pytest

from leapp.libraries.actor.systemfacts import _get_system_groups, _get_system_users, anyendswith, anyhasprefix, aslist
from leapp.libraries.common.testutils import logger_mocked
from leapp.libraries.stdlib import api
from leapp.snactor.fixture import current_actor_libraries


def test_anyendswith(current_actor_libraries):
    value = 'this_is_a_test'

    assert anyendswith(value, ['a_test'])
    assert anyendswith(value, ['a_test', 'bgerwh', 'g52h4q'])
    assert anyendswith(value, ['est'])
    assert not anyendswith(value, ['asafsaf', 'gbfdshh', '123f', 'gdsgsnb'])
    assert not anyendswith(value, [])


def test_anyhasprefix(current_actor_libraries):
    value = 'this_is_a_test'

    assert anyhasprefix(value, ['this'])
    assert anyhasprefix(value, ['this', 'ddsvssd', 'bsdhn', '125fff'])
    assert anyhasprefix(value, ['this_is'])
    assert not anyhasprefix(value, ['ccbbb', 'xasbnn', 'xavavav', 'bbnkk1'])
    assert not anyhasprefix(value, [])


def test_aslist(current_actor_libraries):

    @aslist
    def local():
        yield True
        yield False
        yield True

    r = local()

    assert isinstance(r, list) and r[0] and r[2] and not r[1]


@pytest.mark.parametrize(
    ('etc_passwd_names', 'etc_passwd_directory', 'skipped_user_names'),
    [
        (['root', 'unbound', 'dbus'], '/', []),
        (['root', '+@scanners', 'dbus', '-@usrc', ''], '/', ['+@scanners', '-@usrc', '']),
        (['root', '+@scanners', 'dbus'], '', ['root', '+@scanners', 'dbus']),
    ]
)
def test_get_system_users(monkeypatch, etc_passwd_names, etc_passwd_directory, skipped_user_names):

    class MockedPwdEntry(object):
        def __init__(self, pw_name, pw_uid, pw_gid, pw_dir):
            self.pw_name = pw_name
            self.pw_uid = pw_uid
            self.pw_gid = pw_gid
            self.pw_dir = pw_dir

    etc_passwd_contents = []
    for etc_passwd_name in etc_passwd_names:
        etc_passwd_contents.append(MockedPwdEntry(etc_passwd_name, 0, 0, etc_passwd_directory))

    monkeypatch.setattr(pwd, 'getpwall', lambda: etc_passwd_contents)
    monkeypatch.setattr(api, 'current_logger', logger_mocked())

    _get_system_users()

    if skipped_user_names:
        assert len(api.current_logger().dbgmsg) == 1

        for skipped_user_name in skipped_user_names:
            assert skipped_user_name in api.current_logger().dbgmsg[0]

        for user_name in etc_passwd_names:
            if user_name not in skipped_user_names:
                assert user_name not in api.current_logger().dbgmsg[0]
    else:
        assert not api.current_logger().dbgmsg


@pytest.mark.parametrize(
    ('etc_group_names', 'skipped_group_names'),
    [
        (['cdrom', 'floppy', 'tape'], []),
        (['cdrom', '+@scanners', 'floppy', '-@usrc', ''], ['+@scanners', '-@usrc', '']),
    ]
)
def test_get_system_groups(monkeypatch, etc_group_names, skipped_group_names):

    class MockedGrpEntry(object):
        def __init__(self, gr_name, gr_gid, gr_mem):
            self.gr_name = gr_name
            self.gr_gid = gr_gid
            self.gr_mem = gr_mem

    etc_group_contents = []
    for etc_group_name in etc_group_names:
        etc_group_contents.append(MockedGrpEntry(etc_group_name, 0, []))

    monkeypatch.setattr(grp, 'getgrall', lambda: etc_group_contents)
    monkeypatch.setattr(api, 'current_logger', logger_mocked())

    _get_system_groups()

    if skipped_group_names:
        assert len(api.current_logger().dbgmsg) == 1

        for skipped_group_name in skipped_group_names:
            assert skipped_group_name in api.current_logger().dbgmsg[0]

        for group_name in etc_group_names:
            if group_name not in skipped_group_names:
                assert group_name not in api.current_logger().dbgmsg[0]
    else:
        assert not api.current_logger().dbgmsg



import pytest

from leapp import reporting
from leapp.libraries.actor import repositoriesblacklist
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import (
    CustomTargetRepository,
    EnvVar,
    PESIDRepositoryEntry,
    RepoMapEntry,
    RepositoriesBlacklisted,
    RepositoriesFacts,
    RepositoriesMapping,
    RepositoryData,
    RepositoryFile
)


@pytest.fixture
def repofacts_opts_disabled():
    repos_data = [
        RepositoryData(
            repoid="rhel-7-server-optional-rpms",
            name="RHEL 7 Server",
            enabled=False,
        )
    ]
    repos_files = [
        RepositoryFile(file="/etc/yum.repos.d/redhat.repo", data=repos_data)
    ]
    return RepositoriesFacts(repositories=repos_files)


@pytest.fixture
def rhel7_optional_pesidrepo():
    return PESIDRepositoryEntry(
        pesid='rhel7-optional',
        major_version='7',
        repoid='rhel-7-server-optional-rpms',
        rhui='',
        arch='x86_64',
        channel='ga',
        repo_type='rpm',
    )


@pytest.fixture
def rhel8_crb_pesidrepo():
    return PESIDRepositoryEntry(
        pesid='rhel8-CRB',
        major_version='8',
        repoid='codeready-builder-for-rhel-8-x86_64-rpms',
        rhui='',
        arch='x86_64',
        channel='ga',
        repo_type='rpm')


@pytest.fixture
def repomap_opts_only(rhel7_optional_pesidrepo, rhel8_crb_pesidrepo):
    return RepositoriesMapping(
        mapping=[RepoMapEntry(source='rhel7-optional', target=['rhel8-CRB'])],
        repositories=[rhel7_optional_pesidrepo, rhel8_crb_pesidrepo]
    )


def test_all_target_optionals_blacklisted_when_no_optional_on_source(monkeypatch, repomap_opts_only):
    """
    Tests whether every target optional repository gets blacklisted
    if no optional repositories are used on the source system.
    """

    repos_data = [
        RepositoryData(
            repoid="rhel-7-server-rpms",
            name="RHEL 7 Server",
            enabled=True,
        )
    ]
    repos_files = [
        RepositoryFile(file="/etc/yum.repos.d/redhat.repo", data=repos_data)
    ]
    repo_facts = RepositoriesFacts(repositories=repos_files)

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=[repo_facts, repomap_opts_only]))
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(reporting, 'create_report', produce_mocked())

    repositoriesblacklist.process()

    assert api.produce.called
    assert 'codeready-builder-for-rhel-8-x86_64-rpms' in api.produce.model_instances[0].repoids


def test_with_no_mapping_for_optional_repos(monkeypatch, repomap_opts_only, repofacts_opts_disabled):
    """
    Tests whether nothing gets produced if no valid target is found for an optional repository in mapping data.
    """

    repomap_opts_only.repositories[1].pesid = 'test_pesid'

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=[repofacts_opts_disabled, repomap_opts_only]))
    monkeypatch.setattr(api, 'produce', produce_mocked())

    repositoriesblacklist.process()

    assert not api.produce.called


def test_blacklist_produced_when_optional_repo_disabled(monkeypatch, repofacts_opts_disabled, repomap_opts_only):
    """
    Tests whether a correct blacklist is generated when there is disabled optional repo on the system.
    """

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=[repofacts_opts_disabled, repomap_opts_only]))
    monkeypatch.setattr(api, "produce", produce_mocked())
    monkeypatch.setattr(reporting, "create_report", produce_mocked())

    repositoriesblacklist.process()

    assert api.produce.model_instances, 'A blacklist should get generated.'

    expected_blacklisted_repoid = 'codeready-builder-for-rhel-8-x86_64-rpms'
    err_msg = 'Blacklist does not contain expected repoid.'
    assert expected_blacklisted_repoid in api.produce.model_instances[0].repoids, err_msg


def test_no_blacklist_produced_when_optional_repo_enabled(monkeypatch, repofacts_opts_disabled, repomap_opts_only):
    """
    Tests whether nothing is produced when an optional repository is enabled.

    Data are set up in such a fashion so that the determined blacklist would not be empty.
    """

    repofacts_opts_disabled.repositories[0].data[0].enabled = True

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=[repofacts_opts_disabled, repomap_opts_only]))
    monkeypatch.setattr(api, "produce", produce_mocked())
    monkeypatch.setattr(reporting, "create_report", produce_mocked())

    repositoriesblacklist.process()

    assert not api.produce.called


def test_repositoriesblacklist_not_empty(monkeypatch, repofacts_opts_disabled, repomap_opts_only):
    """
    Tests whether a message containing correct packages from the determined blacklist is produced.
    """

    blacklisted_repoid = 'test'
    monkeypatch.setattr(repositoriesblacklist, "_get_repoids_to_exclude", lambda dummy_mapping: {blacklisted_repoid})
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=[repofacts_opts_disabled, repomap_opts_only]))
    monkeypatch.setattr(api, "produce", produce_mocked())
    monkeypatch.setattr(reporting, "create_report", produce_mocked())

    repositoriesblacklist.process()
    assert api.produce.called == 1
    assert isinstance(api.produce.model_instances[0], RepositoriesBlacklisted)
    assert api.produce.model_instances[0].repoids[0] == blacklisted_repoid
    assert reporting.create_report.called == 1


def test_repositoriesblacklist_empty(monkeypatch, repofacts_opts_disabled, repomap_opts_only):
    """
    Tests whether nothing is produced if there are some disabled optional repos, but an empty blacklist is determined
    from the repo mapping data.
    """

    msgs_to_feed = [repofacts_opts_disabled, repomap_opts_only]

    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=msgs_to_feed))
    monkeypatch.setattr(
        repositoriesblacklist,
        "_get_repoids_to_exclude",
        lambda dummy_mapping: set()
    )  # pylint:disable=W0108
    monkeypatch.setattr(api, "produce", produce_mocked())

    repositoriesblacklist.process()
    assert api.produce.called == 0


@pytest.mark.parametrize(
    ("enabled_repo", "exp_report_title", "message_produced"),
    [
        ("codeready-builder-for-rhel-8-x86_64-rpms", "Using repository not supported by Red Hat", False),
        ("some_other_enabled_repo", "Excluded target system repositories", True),
        (None, "Excluded target system repositories", True),
    ],
)
def test_enablerepo_option(monkeypatch,
                           repofacts_opts_disabled,
                           repomap_opts_only,
                           enabled_repo,
                           exp_report_title,
                           message_produced):
    """
    Tests whether the actor respects CustomTargetRepository messages when constructing the blacklist.
    """

    msgs_to_feed = [repomap_opts_only, repofacts_opts_disabled]

    if enabled_repo:
        msgs_to_feed.append(CustomTargetRepository(repoid=enabled_repo))
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(msgs=msgs_to_feed))
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())
    repositoriesblacklist.process()
    assert reporting.create_report.report_fields["title"] == exp_report_title
    if message_produced:
        assert isinstance(api.produce.model_instances[0], RepositoriesBlacklisted)
    else:
        assert not api.produce.model_instances



import os

import pytest

from leapp.libraries.actor import scancpu
from leapp.libraries.common import testutils
from leapp.libraries.common.config.architecture import (
    ARCH_ARM64,
    ARCH_PPC64LE,
    ARCH_S390X,
    ARCH_SUPPORTED,
    ARCH_X86_64
)
from leapp.libraries.stdlib import api
from leapp.models import CPUInfo

CUR_DIR = os.path.dirname(os.path.abspath(__file__))

LSCPU = {
    ARCH_ARM64: {
        "machine_type": None,
        "flags": ['fp', 'asimd', 'evtstrm', 'aes', 'pmull', 'sha1', 'sha2', 'crc32', 'cpuid'],
    },
    ARCH_PPC64LE: {
        "machine_type": None,
        "flags": []
    },
    ARCH_S390X: {
        "machine_type":
            2827,
        "flags": [
            'esan3', 'zarch', 'stfle', 'msa', 'ldisp', 'eimm', 'dfp', 'edat', 'etf3eh', 'highgprs', 'te', 'vx', 'vxd',
            'vxe', 'gs', 'vxe2', 'vxp', 'sort', 'dflt', 'sie'
        ]
    },
    ARCH_X86_64: {
        "machine_type":
            None,
        "flags": [
            'fpu', 'vme', 'de', 'pse', 'tsc', 'msr', 'pae', 'mce', 'cx8', 'apic', 'sep', 'mtrr', 'pge', 'mca', 'cmov',
            'pat', 'pse36', 'clflush', 'dts', 'acpi', 'mmx', 'fxsr', 'sse', 'sse2', 'ss', 'ht', 'tm', 'pbe', 'syscall',
            'nx', 'pdpe1gb', 'rdtscp', 'lm', 'constant_tsc', 'arch_perfmon', 'pebs', 'bts', 'rep_good', 'nopl',
            'xtopology', 'nonstop_tsc', 'cpuid', 'aperfmperf', 'pni', 'pclmulqdq', 'dtes64', 'monitor', 'ds_cpl',
            'vmx', 'smx', 'est', 'tm2', 'ssse3', 'sdbg', 'fma', 'cx16', 'xtpr', 'pdcm', 'pcid', 'dca', 'sse4_1',
            'sse4_2', 'x2apic', 'movbe', 'popcnt', 'tsc_deadline_timer', 'aes', 'xsave', 'avx', 'f16c', 'rdrand',
            'lahf_lm', 'abm', 'cpuid_fault', 'epb', 'invpcid_single', 'pti', 'ssbd', 'ibrs', 'ibpb', 'stibp',
            'tpr_shadow', 'vnmi', 'flexpriority', 'ept', 'vpid', 'ept_ad', 'fsgsbase', 'tsc_adjust', 'bmi1', 'avx2',
            'smep', 'bmi2', 'erms', 'invpcid', 'cqm', 'xsaveopt', 'cqm_llc', 'cqm_occup_llc', 'dtherm', 'ida', 'arat',
            'pln', 'pts', 'md_clear', 'flush_l1d'
        ]
    },
}


class mocked_get_cpuinfo(object):

    def __init__(self, filename):
        self.filename = filename

    def __call__(self):
        """
        Return lines of the self.filename test file located in the files directory.

        Those files contain /proc/cpuinfo content from several machines.
        """
        with open(os.path.join(CUR_DIR, 'files', self.filename), 'r') as fp:
            return '\n'.join(fp.read().splitlines())


@pytest.mark.parametrize("arch", ARCH_SUPPORTED)
def test_scancpu(monkeypatch, arch):

    mocked_cpuinfo = mocked_get_cpuinfo('lscpu_' + arch)
    monkeypatch.setattr(scancpu, '_get_lscpu_output', mocked_cpuinfo)
    monkeypatch.setattr(api, 'produce', testutils.produce_mocked())
    current_actor = testutils.CurrentActorMocked(arch=arch)
    monkeypatch.setattr(api, 'current_actor', current_actor)

    scancpu.process()

    expected = CPUInfo(machine_type=LSCPU[arch]["machine_type"], flags=LSCPU[arch]["flags"])
    produced = api.produce.model_instances[0]

    assert api.produce.called == 1

    # Produced what was expected
    assert expected.machine_type == produced.machine_type
    assert sorted(expected.flags) == sorted(produced.flags)

    # Did not produce anything extra
    assert expected == produced



from collections import namedtuple
from os import access, path, symlink, X_OK

import pytest

from leapp.libraries.actor import workaround
from leapp.libraries.common.utils import makedirs

SysVersionInfo = namedtuple('version_info', ['major', 'minor', 'micro', 'releaselevel', 'serial'])


def fake_symlink(basedir):
    def impl(source, target):
        source_path = str(basedir.join(*source.lstrip('/').split('/')))
        makedirs(source_path)
        symlink(source_path, target)
    return impl


def test_apply_python3_workaround(monkeypatch, tmpdir):
    leapp_home = tmpdir.mkdir('tmp_leapp_py3')
    monkeypatch.setattr(workaround.os, 'symlink', fake_symlink(tmpdir.mkdir('lib')))
    monkeypatch.setattr(workaround, 'LEAPP_HOME', str(leapp_home))

    # Ensure double invocation doesn't cause a problem
    workaround.apply_python3_workaround()
    workaround.apply_python3_workaround()

    # Ensure creation of all required elements
    assert path.islink(str(leapp_home.join('leapp')))
    assert path.isfile(str(leapp_home.join('leapp3')))
    assert access(str(leapp_home.join('leapp3')), X_OK)

    assert str(leapp_home) in leapp_home.join('leapp3').read_text('utf-8')


@pytest.mark.parametrize('pydir', ('python2.7', 'python3.6', 'python3.9'))
def test_orig_leapp_path(monkeypatch, pydir):
    monkeypatch.setattr(workaround, '_get_python_dirname', lambda: pydir)
    assert workaround._get_orig_leapp_path() == '/usr/lib/{}/site-packages/leapp'.format(pydir)


@pytest.mark.parametrize('sys_version_info,result', (
    (SysVersionInfo(2, 7, 5, 'final', 0), 'python2.7'),
    (SysVersionInfo(3, 6, 0, 'X', 0), 'python3.6'),
    (SysVersionInfo(3, 9, 0, 'X', 0), 'python3.9'),
))
def test_get_python_dirname(monkeypatch, sys_version_info, result):
    monkeypatch.setattr(workaround.sys, 'version_info', sys_version_info)
    assert workaround._get_python_dirname() == result



from leapp import reporting
from leapp.reporting import create_report


def _merge_configs(configs):
    options = {'default_path_checker': None, 'detect_prio': None,
               'detect_path_checker': None, 'reassign_maps': None,
               'retain_attached_hw_handler': None}
    for config in configs:
        if config.default_path_checker is not None:
            options['default_path_checker'] = (config.default_path_checker,
                                               config.pathname)

        if config.reassign_maps is not None:
            options['reassign_maps'] = (config.reassign_maps, config.pathname)

        if config.default_detect_checker is not None:
            options['detect_path_checker'] = (config.default_detect_checker,
                                              config.pathname)

        if config.default_detect_prio is not None:
            options['detect_prio'] = (config.default_detect_prio,
                                      config.pathname)

        if config.default_retain_hwhandler is not None:
            options['retain_attached_hw_handler'] = (config.default_retain_hwhandler, config.pathname)
    return options


def _check_default_path_checker(options):
    if not options['default_path_checker']:
        return
    value, pathname = options['default_path_checker']
    if value == 'tur':
        return
    create_report([
        reporting.Title(
            'Unsupported device-mapper-multipath configuration'
        ),
        reporting.Summary(
            'device-mapper-multipath has changed the default path_checker '
            'from "directio" to "tur" in RHEL-8. Further, changing the '
            'default path_checker can cause issues with built-in device '
            'configurations in RHEL-8. Please remove the "path_checker" '
            'option from the defaults section of {}, and add it to the '
            'device configuration of any devices that need it.'.
            format(pathname)
        ),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.SERVICES]),
        reporting.Groups([reporting.Groups.INHIBITOR]),
        reporting.RelatedResource('package', 'device-mapper-multipath'),
        reporting.RelatedResource('file', pathname),
        reporting.Remediation(
            hint='Please remove the "path_checker {}" option from the '
            'defaults section of {}, and add it to the device configuration '
            'of any devices that need it.'.format(value, pathname)
        )
    ])


def _create_paths_str(paths):
    if len(paths) < 2:
        return paths[0]
    return '{} and {}'.format(', '.join(paths[0:-1]), paths[-1])


def _check_default_detection(options):
    bad = []
    for keyword in ('detect_path_checker', 'detect_prio',
                    'retain_attached_hw_handler'):
        if options[keyword] and not options[keyword][0] and \
                options[keyword][1] not in bad:
            bad.append(options[keyword][1])
    if not bad:
        return
    paths = _create_paths_str(bad)
    create_report([
        reporting.Title(
            'device-mapper-multipath now defaults to detecting settings'
        ),
        reporting.Summary(
            'In RHEL-8, the default value for the "detect_path_checker", '
            '"detect_prio" and "retain_attached_hw_handler" options has '
            'changed to "yes". Further, changing these default values can '
            'cause issues with the built-in device configurations in RHEL-8. '
            'They will be commented out in the defaults section of all '
            'multipath config files. This is unlikely to cause any issues '
            'with existing configurations. If it does, please move these '
            'options from the defaults sections of {} to the device '
            'configuration sections of any devices that need them.'.
            format(paths)
        ),
        reporting.Severity(reporting.Severity.MEDIUM),
        reporting.Groups([reporting.Groups.SERVICES]),
        reporting.RelatedResource('package', 'device-mapper-multipath')
    ])


def _check_reassign_maps(options):
    if not options['reassign_maps']:
        return
    value, pathname = options['reassign_maps']
    if not value:
        return
    create_report([
        reporting.Title(
            'device-mapper-multipath now disables reassign_maps by default'
        ),
        reporting.Summary(
            'In RHEL-8, the default value for "reassign_maps" has been '
            'changed to "no", and it is not recommended to enable it in any '
            'configuration going forward. This option will be commented out '
            'in {}.'.format(pathname)
        ),
        reporting.Severity(reporting.Severity.MEDIUM),
        reporting.Groups([reporting.Groups.SERVICES]),
        reporting.RelatedResource('package', 'device-mapper-multipath')
    ])


def check_configs(facts):
    options = _merge_configs(facts.configs)
    _check_default_path_checker(options)
    _check_default_detection(options)
    _check_reassign_maps(options)



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class CupsChangedFeatures(Model):
    topic = SystemInfoTopic

    interface = fields.Boolean(default=False)
    """
    True if interface scripts are used, False otherwise
    """

    digest = fields.Boolean(default=False)
    """
    True if Digest/BasicDigest directive values are used, False otherwise
    """

    include = fields.Boolean(default=False)
    """
    True if Include directive is used, False otherwise
    """

    certkey = fields.Boolean(default=False)
    """
    True if ServerKey/ServerCertificate directives are used, False otherwise
    """

    env = fields.Boolean(default=False)
    """
    True if PassEnv/SetEnv directives are used, False otherwise
    """

    printcap = fields.Boolean(default=False)
    """
    True if PrintcapFormat directive is used, False otherwise
    """

    include_files = fields.List(fields.String(), default=['/etc/cups/cupsd.conf'])
    """
    Paths to included files, contains /etc/cups/cupsd.conf by default
    """



import os
from collections import namedtuple

import six

from leapp.libraries.common.config import architecture as arch
from leapp.libraries.common.config.version import get_source_major_version, get_target_major_version
from leapp.libraries.stdlib import api
from leapp.utils.deprecation import deprecated

# when on AWS and upgrading from RHEL 7, we need also Python2 version of "Amazon-id" dnf
# plugin which is served by "leapp-rhui-aws" rpm package (please note this package is not
# in any RH official repository but only in "rhui-client-config-*" repo)
DNF_PLUGIN_PATH_PY2 = '/usr/lib/python2.7/site-packages/dnf-plugins/'
YUM_REPOS_PATH = '/etc/yum.repos.d'

RHUI_PKI_DIR = '/etc/pki/rhui'
RHUI_PKI_PRODUCT_DIR = os.path.join(RHUI_PKI_DIR, 'product')
RHUI_PKI_PRIVATE_DIR = os.path.join(RHUI_PKI_DIR, 'private')

AWS_DNF_PLUGIN_NAME = 'amazon-id.py'


class ContentChannel(object):
    GA = 'ga'
    TUV = 'tuv'
    E4S = 'e4s'
    EUS = 'eus'
    AUS = 'aus'
    BETA = 'beta'


class RHUIVariant(object):
    ORDINARY = 'ordinary'  # Special value - not displayed in report/errors
    SAP = 'sap'
    SAP_APPS = 'sap-apps'
    SAP_HA = 'sap-ha'


class RHUIProvider(object):
    GOOGLE = 'Google'
    AZURE = 'Azure'
    AWS = 'AWS'
    ALIBABA = 'Alibaba'


# The files in 'files_map' are provided by special Leapp rpms (per cloud) and
# are delivered into "repos/system_upgrade/common/files/rhui/<PROVIDER>

RHUISetup = namedtuple(
    'RHUISetup',
    ('clients', 'leapp_pkg', 'mandatory_files', 'optional_files', 'extra_info', 'os_version',
     'arch', 'content_channel', 'files_supporting_client_operation')
)
"""RHUI-Setup-specific details used during IPU
.. py:attribute:: clients
    A set of RHUI clients present on the system.
.. py:attribute:: leapp_pkg
    The name of leapp's rhui-specific pkg providing repofiles, certs and keys to access package of the setup.
.. py:attribute:: mandatory_files
    Mandatory files and their destinations to copy into target userspace container required to access the target OS
    content. If not present, an exception will be raised.
.. py:attribute:: optional_files
    Optional files and their destinations to copy into target userspace container required to access the target OS
    content. Nonexistence of any of these files is ignored.
.. py:attribute:: extra_info
    Extra information about the setup.
.. py:attribute:: os_version
    The major OS version of the RHUI system.
.. py:attribute:: content_channel
    Content channel used by the RHUI setup.
.. py:attribute:: files_supporting_client_operation
    A subset of files from ``mandatory_files`` that are necessary for client to work (cannot be cleaned up).
"""


class RHUIFamily(object):
    def __init__(self, provider, client_files_folder='', variant=RHUIVariant.ORDINARY, arch=arch.ARCH_X86_64,):
        self.provider = provider
        self.client_files_folder = client_files_folder
        self.variant = variant
        self.arch = arch

    def __hash__(self):
        return hash((self.provider, self.variant, self.arch))

    def __eq__(self, other):
        if not isinstance(other, RHUIFamily):
            return False
        self_repr = (self.provider, self.variant, self.arch)
        other_repr = (other.provider, other.variant, other.arch)
        return self_repr == other_repr

    def full_eq(self, other):
        partial_eq = self == other
        return partial_eq and self.client_files_folder == other.client_files_folder

    def __str__(self):
        template = 'RHUIFamily(provider={provider}, variant={variant}, arch={arch})'
        return template.format(provider=self.provider, variant=self.variant, arch=self.arch)


def mk_rhui_setup(clients=None, leapp_pkg='', mandatory_files=None, optional_files=None,
                  extra_info=None, os_version='7', arch=arch.ARCH_X86_64, content_channel=ContentChannel.GA,
                  files_supporting_client_operation=None):
    clients = clients or set()
    mandatory_files = mandatory_files or []
    extra_info = extra_info or {}
    files_supporting_client_operation = files_supporting_client_operation or []

    # Since the default optional files are not [], we cannot use the same construction as above
    # to allow the caller to specify empty optional files
    default_opt_files = [('content-leapp.crt', RHUI_PKI_PRODUCT_DIR), ('key-leapp.pem', RHUI_PKI_DIR)]
    optional_files = default_opt_files if optional_files is None else optional_files

    return RHUISetup(clients=clients, leapp_pkg=leapp_pkg, mandatory_files=mandatory_files, arch=arch,
                     content_channel=content_channel, optional_files=optional_files, extra_info=extra_info,
                     os_version=os_version, files_supporting_client_operation=files_supporting_client_operation)


# This will be the new "cloud map". Essentially a directed graph with edges defined implicitly by OS versions +
# setup family identification. In theory, we can make the variant be part of rhui setups, but this way we don't
# have to repeatedly write it to every known setup there is (a sort of compression). Furthermore, it limits
# the search for target equivalent to setups sharing the same family, and thus reducing a chance of error.
RHUI_SETUPS = {
    RHUIFamily(RHUIProvider.AWS, client_files_folder='aws'): [
        mk_rhui_setup(clients={'rh-amazon-rhui-client'}, optional_files=[], os_version='7'),
        mk_rhui_setup(clients={'rh-amazon-rhui-client'}, leapp_pkg='leapp-rhui-aws',
                      mandatory_files=[
                        ('rhui-client-config-server-8.crt', RHUI_PKI_PRODUCT_DIR),
                        ('rhui-client-config-server-8.key', RHUI_PKI_DIR),
                        (AWS_DNF_PLUGIN_NAME, DNF_PLUGIN_PATH_PY2),
                        ('leapp-aws.repo', YUM_REPOS_PATH)
                      ],
                      files_supporting_client_operation=[AWS_DNF_PLUGIN_NAME],
                      optional_files=[
                        ('content-rhel8.key', RHUI_PKI_DIR),
                        ('cdn.redhat.com-chain.crt', RHUI_PKI_DIR),
                        ('content-rhel8.crt', RHUI_PKI_PRODUCT_DIR)
                      ], os_version='8'),
        # @Note(mhecko): We don't need to deal with AWS_DNF_PLUGIN_NAME here as on rhel8+ we can use the plugin
        # #              provided by the target client - there is no Python2 incompatibility issue there.
        mk_rhui_setup(clients={'rh-amazon-rhui-client'}, leapp_pkg='leapp-rhui-aws',
                      mandatory_files=[
                        ('rhui-client-config-server-9.crt', RHUI_PKI_PRODUCT_DIR),
                        ('rhui-client-config-server-9.key', RHUI_PKI_DIR),
                        ('leapp-aws.repo', YUM_REPOS_PATH)
                      ],
                      optional_files=[
                        ('content-rhel9.key', RHUI_PKI_DIR),
                        ('cdn.redhat.com-chain.crt', RHUI_PKI_DIR),
                        ('content-rhel9.crt', RHUI_PKI_PRODUCT_DIR)
                      ], os_version='9'),
    ],
    RHUIFamily(RHUIProvider.AWS, arch=arch.ARCH_ARM64, client_files_folder='aws'): [
        mk_rhui_setup(clients={'rh-amazon-rhui-client-arm'}, optional_files=[], os_version='7', arch=arch.ARCH_ARM64),
        mk_rhui_setup(clients={'rh-amazon-rhui-client-arm'}, leapp_pkg='leapp-rhui-aws',
                      mandatory_files=[
                        ('rhui-client-config-server-8.crt', RHUI_PKI_PRODUCT_DIR),
                        ('rhui-client-config-server-8.key', RHUI_PKI_DIR),
                        (AWS_DNF_PLUGIN_NAME, DNF_PLUGIN_PATH_PY2),
                        ('leapp-aws.repo', YUM_REPOS_PATH)
                      ],
                      files_supporting_client_operation=[AWS_DNF_PLUGIN_NAME],
                      optional_files=[
                        ('content-rhel8.key', RHUI_PKI_DIR),
                        ('cdn.redhat.com-chain.crt', RHUI_PKI_DIR),
                        ('content-rhel8.crt', RHUI_PKI_PRODUCT_DIR)
                      ], os_version='8', arch=arch.ARCH_ARM64),
        mk_rhui_setup(clients={'rh-amazon-rhui-client-arm'}, leapp_pkg='leapp-rhui-aws',
                      mandatory_files=[
                        ('rhui-client-config-server-9.crt', RHUI_PKI_PRODUCT_DIR),
                        ('rhui-client-config-server-9.key', RHUI_PKI_DIR),
                        ('leapp-aws.repo', YUM_REPOS_PATH)
                      ],
                      optional_files=[
                        ('content-rhel9.key', RHUI_PKI_DIR),
                        ('cdn.redhat.com-chain.crt', RHUI_PKI_DIR),
                        ('content-rhel9.crt', RHUI_PKI_PRODUCT_DIR)
                      ], os_version='9', arch=arch.ARCH_ARM64),
    ],
    RHUIFamily(RHUIProvider.AWS, variant=RHUIVariant.SAP, client_files_folder='aws-sap-e4s'): [
        mk_rhui_setup(clients={'rh-amazon-rhui-client-sap-bundle'}, optional_files=[], os_version='7',
                      content_channel=ContentChannel.E4S),
        mk_rhui_setup(clients={'rh-amazon-rhui-client-sap-bundle-e4s'}, leapp_pkg='leapp-rhui-aws-sap-e4s',
                      mandatory_files=[
                        ('rhui-client-config-server-8-sap-bundle.crt', RHUI_PKI_PRODUCT_DIR),
                        ('rhui-client-config-server-8-sap-bundle.key', RHUI_PKI_DIR),
                        (AWS_DNF_PLUGIN_NAME, DNF_PLUGIN_PATH_PY2),
                        ('leapp-aws-sap-e4s.repo', YUM_REPOS_PATH)
                      ],
                      files_supporting_client_operation=[AWS_DNF_PLUGIN_NAME],
                      optional_files=[
                        ('content-rhel8-sap.key', RHUI_PKI_DIR),
                        ('cdn.redhat.com-chain.crt', RHUI_PKI_DIR),
                        ('content-rhel8-sap.crt', RHUI_PKI_PRODUCT_DIR)
                      ], os_version='8', content_channel=ContentChannel.E4S),
        mk_rhui_setup(clients={'rh-amazon-rhui-client-sap-bundle-e4s'}, leapp_pkg='leapp-rhui-aws-sap-e4s',
                      mandatory_files=[
                        ('rhui-client-config-server-9-sap-bundle.crt', RHUI_PKI_PRODUCT_DIR),
                        ('rhui-client-config-server-9-sap-bundle.key', RHUI_PKI_DIR),
                        ('leapp-aws-sap-e4s.repo', YUM_REPOS_PATH)
                      ],
                      optional_files=[
                        ('content-rhel9-sap-bundle-e4s.key', RHUI_PKI_DIR),
                        ('cdn.redhat.com-chain.crt', RHUI_PKI_DIR),
                        ('content-rhel9-sap-bundle-e4s.crt', RHUI_PKI_PRODUCT_DIR)
                      ], os_version='9', content_channel=ContentChannel.E4S),
    ],
    RHUIFamily(RHUIProvider.AZURE, client_files_folder='azure'): [
        mk_rhui_setup(clients={'rhui-azure-rhel7'}, os_version='7',
                      extra_info={'agent_pkg': 'WALinuxAgent'}),
        mk_rhui_setup(clients={'rhui-azure-rhel8'}, leapp_pkg='leapp-rhui-azure',
                      mandatory_files=[('leapp-azure.repo', YUM_REPOS_PATH)],
                      optional_files=[
                        ('key.pem', RHUI_PKI_DIR),
                        ('content.crt', RHUI_PKI_PRODUCT_DIR)
                      ],
                      extra_info={'agent_pkg': 'WALinuxAgent'},
                      os_version='8'),
        mk_rhui_setup(clients={'rhui-azure-rhel9'}, leapp_pkg='leapp-rhui-azure',
                      mandatory_files=[('leapp-azure.repo', YUM_REPOS_PATH)],
                      optional_files=[
                        ('key.pem', RHUI_PKI_DIR),
                        ('content.crt', RHUI_PKI_PRODUCT_DIR)
                      ],
                      extra_info={'agent_pkg': 'WALinuxAgent'},
                      os_version='9'),
    ],
    RHUIFamily(RHUIProvider.AZURE, variant=RHUIVariant.SAP_APPS, client_files_folder='azure-sap-apps'): [
        mk_rhui_setup(clients={'rhui-azure-rhel7-base-sap-apps'}, os_version='7', content_channel=ContentChannel.EUS),
        mk_rhui_setup(clients={'rhui-azure-rhel8-sapapps'}, leapp_pkg='leapp-rhui-azure-sap',
                      mandatory_files=[('leapp-azure-sap-apps.repo', YUM_REPOS_PATH)],
                      optional_files=[
                        ('key-sapapps.pem', RHUI_PKI_DIR),
                        ('content-sapapps.crt', RHUI_PKI_PRODUCT_DIR)
                      ],
                      extra_info={'agent_pkg': 'WALinuxAgent'},
                      os_version='8', content_channel=ContentChannel.EUS),
        mk_rhui_setup(clients={'rhui-azure-rhel9-sapapps'}, leapp_pkg='leapp-rhui-azure-sap',
                      mandatory_files=[('leapp-azure-sap-apps.repo', YUM_REPOS_PATH)],
                      optional_files=[
                        ('key-sapapps.pem', RHUI_PKI_DIR),
                        ('content-sapapps.crt', RHUI_PKI_PRODUCT_DIR)
                      ],
                      extra_info={'agent_pkg': 'WALinuxAgent'},
                      os_version='9', content_channel=ContentChannel.EUS),
    ],
    RHUIFamily(RHUIProvider.AZURE, variant=RHUIVariant.SAP_HA, client_files_folder='azure-sap-ha'): [
        mk_rhui_setup(clients={'rhui-azure-rhel7-base-sap-ha'}, os_version='7', content_channel=ContentChannel.E4S),
        mk_rhui_setup(clients={'rhui-azure-rhel8-sap-ha'}, leapp_pkg='leapp-rhui-azure-sap',
                      mandatory_files=[('leapp-azure-sap-ha.repo', YUM_REPOS_PATH)],
                      optional_files=[
                        ('key-sap-ha.pem', RHUI_PKI_DIR),
                        ('content-sap-ha.crt', RHUI_PKI_PRODUCT_DIR)
                      ],
                      extra_info={'agent_pkg': 'WALinuxAgent'},
                      os_version='8', content_channel=ContentChannel.E4S),
        mk_rhui_setup(clients={'rhui-azure-rhel9-sap-ha'}, leapp_pkg='leapp-rhui-azure-sap',
                      mandatory_files=[('leapp-azure-sap-ha.repo', YUM_REPOS_PATH)],
                      optional_files=[
                        ('key-sap-ha.pem', RHUI_PKI_DIR),
                        ('content-sap-ha.crt', RHUI_PKI_PRODUCT_DIR)
                      ],
                      extra_info={'agent_pkg': 'WALinuxAgent'},
                      os_version='9', content_channel=ContentChannel.E4S),
    ],
    RHUIFamily(RHUIProvider.GOOGLE, client_files_folder='google'): [
        mk_rhui_setup(clients={'google-rhui-client-rhel7'}, os_version='7'),
        mk_rhui_setup(clients={'google-rhui-client-rhel8'}, leapp_pkg='leapp-rhui-google',
                      mandatory_files=[('leapp-google.repo', YUM_REPOS_PATH)],
                      files_supporting_client_operation=['leapp-google.repo'],
                      os_version='8'),
        mk_rhui_setup(clients={'google-rhui-client-rhel9'}, leapp_pkg='leapp-rhui-google',
                      mandatory_files=[('leapp-google.repo', YUM_REPOS_PATH)],
                      files_supporting_client_operation=['leapp-google.repo'],
                      os_version='9'),
    ],
    RHUIFamily(RHUIProvider.GOOGLE, variant=RHUIVariant.SAP, client_files_folder='google-sap'): [
        mk_rhui_setup(clients={'google-rhui-client-rhel79-sap'}, os_version='7', content_channel=ContentChannel.E4S),
        mk_rhui_setup(clients={'google-rhui-client-rhel8-sap'}, leapp_pkg='leapp-rhui-google-sap',
                      mandatory_files=[('leapp-google-sap.repo', YUM_REPOS_PATH)],
                      files_supporting_client_operation=['leapp-google-sap.repo'],
                      os_version='8', content_channel=ContentChannel.E4S),
        mk_rhui_setup(clients={'google-rhui-client-rhel9-sap'}, leapp_pkg='leapp-rhui-google-sap',
                      mandatory_files=[('leapp-google-sap.repo', YUM_REPOS_PATH)],
                      files_supporting_client_operation=['leapp-google-sap.repo'],
                      os_version='9', content_channel=ContentChannel.E4S),
    ],
    RHUIFamily(RHUIProvider.ALIBABA, client_files_folder='alibaba'): [
        mk_rhui_setup(clients={'client-rhel7'}, os_version='7'),
        mk_rhui_setup(clients={'aliyun_rhui_rhel8'}, leapp_pkg='leapp-rhui-alibaba',
                      mandatory_files=[('leapp-alibaba.repo', YUM_REPOS_PATH)], os_version='8'),
    ]
}


# DEPRECATED, use RHUI_SETUPS instead
RHUI_CLOUD_MAP = {
    '7to8': {
        'aws': {
            'src_pkg': 'rh-amazon-rhui-client',
            'target_pkg': 'rh-amazon-rhui-client',
            'leapp_pkg': 'leapp-rhui-aws',
            'leapp_pkg_repo': 'leapp-aws.repo',
            'files_map': [
                ('rhui-client-config-server-8.crt', RHUI_PKI_PRODUCT_DIR),
                ('rhui-client-config-server-8.key', RHUI_PKI_DIR),
                ('cdn.redhat.com-chain.crt', RHUI_PKI_DIR),
                (AWS_DNF_PLUGIN_NAME, DNF_PLUGIN_PATH_PY2),
                ('leapp-aws.repo', YUM_REPOS_PATH)
            ],
        },
        'aws-sap-e4s': {
            'src_pkg': 'rh-amazon-rhui-client-sap-bundle',
            'target_pkg': 'rh-amazon-rhui-client-sap-bundle-e4s',
            'leapp_pkg': 'leapp-rhui-aws-sap-e4s',
            'leapp_pkg_repo': 'leapp-aws-sap-e4s.repo',
            'files_map': [
                ('rhui-client-config-server-8-sap-bundle.crt', RHUI_PKI_PRODUCT_DIR),
                ('rhui-client-config-server-8-sap-bundle.key', RHUI_PKI_DIR),
                ('cdn.redhat.com-chain.crt', RHUI_PKI_DIR),
                (AWS_DNF_PLUGIN_NAME, DNF_PLUGIN_PATH_PY2),
                ('leapp-aws-sap-e4s.repo', YUM_REPOS_PATH)
            ],
        },
        'azure': {
            'src_pkg': 'rhui-azure-rhel7',
            'target_pkg': 'rhui-azure-rhel8',
            'agent_pkg': 'WALinuxAgent',
            'leapp_pkg': 'leapp-rhui-azure',
            'leapp_pkg_repo': 'leapp-azure.repo',
            'files_map': [
                ('leapp-azure.repo', YUM_REPOS_PATH)
            ],
        },
        'azure-sap-apps': {
            'src_pkg': 'rhui-azure-rhel7-base-sap-apps',
            'target_pkg': 'rhui-azure-rhel8-sapapps',
            'agent_pkg': 'WALinuxAgent',
            'leapp_pkg': 'leapp-rhui-azure-sap',
            'leapp_pkg_repo': 'leapp-azure-sap-apps.repo',
            'files_map': [
                ('leapp-azure-sap-apps.repo', YUM_REPOS_PATH),
            ],
        },
        'azure-sap-ha': {
            'src_pkg': 'rhui-azure-rhel7-base-sap-ha',
            'target_pkg': 'rhui-azure-rhel8-sap-ha',
            'agent_pkg': 'WALinuxAgent',
            'leapp_pkg': 'leapp-rhui-azure-sap',
            'leapp_pkg_repo': 'leapp-azure-sap-ha.repo',
            'files_map': [
                ('leapp-azure-sap-ha.repo', YUM_REPOS_PATH)
            ],
        },
        'google': {
            'src_pkg': 'google-rhui-client-rhel7',
            'target_pkg': 'google-rhui-client-rhel8',
            'leapp_pkg': 'leapp-rhui-google',
            'leapp_pkg_repo': 'leapp-google.repo',
            'files_map': [
                ('content.crt', RHUI_PKI_PRODUCT_DIR),
                ('key.pem', RHUI_PKI_DIR),
                ('leapp-google.repo', YUM_REPOS_PATH)
            ],
        },
        'google-sap': {
            'src_pkg': 'google-rhui-client-rhel79-sap',
            'target_pkg': 'google-rhui-client-rhel8-sap',
            'leapp_pkg': 'leapp-rhui-google-sap',
            'leapp_pkg_repo': 'leapp-google-sap.repo',
            'files_map': [
                ('content.crt', RHUI_PKI_PRODUCT_DIR),
                ('key.pem', RHUI_PKI_DIR),
                ('leapp-google-sap.repo', YUM_REPOS_PATH)
            ],
        },
        'alibaba': {
            'src_pkg': 'client-rhel7',
            'target_pkg': 'aliyun_rhui_rhel8',
            'leapp_pkg': 'leapp-rhui-alibaba',
            'leapp_pkg_repo': 'leapp-alibaba.repo',
            'files_map': [
                ('content.crt', RHUI_PKI_PRODUCT_DIR),
                ('key.pem', RHUI_PKI_DIR),
                ('leapp-alibaba.repo', YUM_REPOS_PATH)
            ],
        }
    },
    '8to9': {
        'aws': {
            'src_pkg': 'rh-amazon-rhui-client',
            'target_pkg': 'rh-amazon-rhui-client',
            'leapp_pkg': 'leapp-rhui-aws',
            'leapp_pkg_repo': 'leapp-aws.repo',
            'files_map': [
                ('rhui-client-config-server-9.crt', RHUI_PKI_PRODUCT_DIR),
                ('rhui-client-config-server-9.key', RHUI_PKI_DIR),
                ('cdn.redhat.com-chain.crt', RHUI_PKI_DIR),
                ('leapp-aws.repo', YUM_REPOS_PATH)
            ],
        },
        'aws-sap-e4s': {
            'src_pkg': 'rh-amazon-rhui-client-sap-bundle-e4s',
            'target_pkg': 'rh-amazon-rhui-client-sap-bundle-e4s',
            'leapp_pkg': 'leapp-rhui-aws-sap-e4s',
            'leapp_pkg_repo': 'leapp-aws-sap-e4s.repo',
            'files_map': [
                ('rhui-client-config-server-9-sap-bundle.crt', RHUI_PKI_PRODUCT_DIR),
                ('rhui-client-config-server-9-sap-bundle.key', RHUI_PKI_DIR),
                ('cdn.redhat.com-chain.crt', RHUI_PKI_DIR),
                ('leapp-aws-sap-e4s.repo', YUM_REPOS_PATH)
            ],
        },
        'azure': {
            'src_pkg': 'rhui-azure-rhel8',
            'target_pkg': 'rhui-azure-rhel9',
            'agent_pkg': 'WALinuxAgent',
            'leapp_pkg': 'leapp-rhui-azure',
            'leapp_pkg_repo': 'leapp-azure.repo',
            'files_map': [
                ('leapp-azure.repo', YUM_REPOS_PATH)
            ],
        },
        # FIXME(mhecko): This entry is identical to the azure one, since we have no EUS content yet, therefore, it
        # #              serves only the purpose of containing the name of rhui client package to correctly detect
        # #              cloud provider. Trying to work around this entry by specifying --channel, will result in
        # #              failures - there is no repomapping for EUS content, and the name of target pkg differs on EUS.
        # #              If the EUS image is available sooner than the 'azure-eus' entry gets modified, the user can
        # #              still upgrade to non-EUS, and switch the newly upgraded system to EUS manually.
        'azure-eus': {
            'src_pkg': 'rhui-azure-rhel8-eus',
            'target_pkg': 'rhui-azure-rhel9',
            'agent_pkg': 'WALinuxAgent',
            'leapp_pkg': 'leapp-rhui-azure-eus',
            'leapp_pkg_repo': 'leapp-azure.repo',
            'files_map': [
                ('leapp-azure.repo', YUM_REPOS_PATH)
            ],
        },
        'azure-sap-ha': {
            'src_pkg': 'rhui-azure-rhel8-sap-ha',
            'target_pkg': 'rhui-azure-rhel9-sap-ha',
            'agent_pkg': 'WALinuxAgent',
            'leapp_pkg': 'leapp-rhui-azure-sap',
            'leapp_pkg_repo': 'leapp-azure-sap-ha.repo',
            'files_map': [
                ('leapp-azure-sap-ha.repo', YUM_REPOS_PATH)
            ],
        },
        'azure-sap-apps': {
            'src_pkg': 'rhui-azure-rhel8-sapapps',
            'target_pkg': 'rhui-azure-rhel9-sapapps',
            'agent_pkg': 'WALinuxAgent',
            'leapp_pkg': 'leapp-rhui-azure-sap',
            'leapp_pkg_repo': 'leapp-azure-sap-apps.repo',
            'files_map': [
                ('leapp-azure-sap-apps.repo', YUM_REPOS_PATH)
            ],
        },
        'google': {
            'src_pkg': 'google-rhui-client-rhel8',
            'target_pkg': 'google-rhui-client-rhel9',
            'leapp_pkg': 'leapp-rhui-google',
            'leapp_pkg_repo': 'leapp-google.repo',
            'files_map': [
                ('content.crt', RHUI_PKI_PRODUCT_DIR),
                ('key.pem', RHUI_PKI_DIR),
                ('leapp-google.repo', YUM_REPOS_PATH)
            ],
        },
        'google-sap': {
            'src_pkg': 'google-rhui-client-rhel8-sap',
            'target_pkg': 'google-rhui-client-rhel9-sap',
            'leapp_pkg': 'leapp-rhui-google-sap',
            'leapp_pkg_repo': 'leapp-google-sap.repo',
            'files_map': [
                ('content.crt', RHUI_PKI_PRODUCT_DIR),
                ('key.pem', RHUI_PKI_DIR),
                ('leapp-google-sap.repo', YUM_REPOS_PATH)
            ],
        },
        'alibaba': {
            'src_pkg': 'aliyun_rhui_rhel8',
            'target_pkg': 'aliyun_rhui_rhel9',
            'leapp_pkg': 'leapp-rhui-alibaba',
            'leapp_pkg_repo': 'leapp-alibaba.repo',
            'files_map': [
                ('content.crt', RHUI_PKI_PRODUCT_DIR),
                ('key.pem', RHUI_PKI_DIR),
                ('leapp-alibaba.repo', YUM_REPOS_PATH)
            ],
        },
    },
}


def get_upg_path():
    """
    Get upgrade path in specific string format
    """
    return '7to8' if get_target_major_version() == '8' else '8to9'


@deprecated(since='2023-07-27', message='This functionality has been replaced with the RHUIInfo message.')
def gen_rhui_files_map():
    """
    Generate RHUI files map based on architecture and upgrade path
    """
    arch = api.current_actor().configuration.architecture
    upg_path = get_upg_path()

    cloud_map = RHUI_CLOUD_MAP
    # for the moment the only arch related difference in RHUI package naming is on ARM
    if arch == 'aarch64':
        cloud_map[get_upg_path()]['aws']['src_pkg'] = 'rh-amazon-rhui-client-arm'

    files_map = dict((k, v['files_map']) for k, v in six.iteritems(cloud_map[upg_path]))
    return files_map


@deprecated(since='2023-07-27', message='This functionality has been integrated into target_userspace_creator.')
def copy_rhui_data(context, provider):
    """
    Copy relevant RHUI certificates and key into the target userspace container
    """
    rhui_dir = api.get_common_folder_path('rhui')
    data_dir = os.path.join(rhui_dir, provider)

    context.call(['mkdir', '-p', RHUI_PKI_PRODUCT_DIR])
    context.call(['mkdir', '-p', RHUI_PKI_PRIVATE_DIR])

    for path_ in gen_rhui_files_map().get(provider, ()):
        context.copy_to(os.path.join(data_dir, path_[0]), path_[1])


def get_all_known_rhui_pkgs_for_current_upg():
    upg_major_versions = (get_source_major_version(), get_target_major_version())

    known_pkgs = set()
    for setup_family in RHUI_SETUPS.values():
        for setup in setup_family:
            if setup.os_version not in upg_major_versions:
                continue
            known_pkgs.update(setup.clients)
            known_pkgs.add(setup.leapp_pkg)

    return known_pkgs



from leapp.actors import Actor
from leapp.libraries.actor import scanclienablerepo
from leapp.models import CustomTargetRepository
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanCLIenablrepo(Actor):
    """
    Produce CustomTargetRepository based on the LEAPP_ENABLE_REPOS in config.
    """

    name = 'scanclienablerepo'
    consumes = ()
    produces = (CustomTargetRepository,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        scanclienablerepo.process()
